(*--- 

frontend: A prototype frontend for probabilistic programs

    Copyright (C) 2013  Aleksandar Chakarov <alch4231@colorado.edu>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

--*)

open Imp
open Util

let concrete_output = ref false

let path_prefix = 10000

type concrete_state = {real_state: (float LocMap.t); int_state: (int LocMap.t)}

let empty_concrete_state: concrete_state = {real_state = LocMap.empty; int_state = LocMap.empty}

let lookup (sigma: concrete_state) (x: loc) : aexp = 
  let r_sigma = sigma.real_state in
  let i_sigma = sigma.int_state in
  let res = 
    try
      RealConst(LocMap.find x r_sigma)
    with Not_found -> (
      try
	IntConst(LocMap.find x i_sigma)
    with Not_found -> failwith ("Concrete lookup for var "^x^" failed."))
  in res

let update (sigma: concrete_state) (x: loc) (n: aexp) : concrete_state = 
  match n with 
    | IntConst(n') -> let i_sigma' = LocMap.add x n' sigma.int_state in
		      {real_state = sigma.real_state; int_state = i_sigma'}
    | RealConst(r') -> let r_sigma' = LocMap.add x r' sigma.real_state in
		       {real_state = r_sigma'; int_state = sigma.int_state}
    | _ -> failwith ("Unexpected expression for var "^x^" in concrete state : "^(aexp_to_str n))


(* Concretely evaluate program *)
let rec eval_aexp (a: aexp) (sigma: concrete_state) : aexp = match a with
  | IntConst n -> IntConst n
  | RealConst r -> RealConst r
  | Var x -> let res = lookup sigma x in
	     res
  | Add (a0,a1) -> 
    let c0  = eval_aexp a0 sigma in
    let c1  = eval_aexp a1 sigma in
    let res = match (c0, c1) with
      | (IntConst(n1),IntConst(n2)) -> IntConst(n1+n2)
      | (RealConst(r1),RealConst(r2)) -> RealConst(r1+.r2)
      | (_,_) -> failwith ("Add aexp in eval_aexp is not well-typed: "^(aexp_to_str a))
    in res
  | Sub (a0,a1) ->
    let c0  = eval_aexp a0 sigma in
    let c1  = eval_aexp a1 sigma in
    let res = match (c0, c1) with
      | (IntConst(n1),IntConst(n2)) -> IntConst(n1-n2)
      | (RealConst(r1),RealConst(r2)) -> RealConst(r1-.r2)
      | (_,_) -> failwith ("Sub aexp in eval_aexp is not well-typed: "^(aexp_to_str a))
    in res
  | Mul (a0,a1) -> 
    let c0  = eval_aexp a0 sigma in
    let c1  = eval_aexp a1 sigma in
    let res = match (c0, c1) with
      | (IntConst(n1),IntConst(n2)) -> IntConst(n1*n2)
      | (RealConst(r1),RealConst(r2)) -> RealConst(r1*.r2)
      | (_,_) -> failwith ("Mul aexp in eval_aexp is not well-typed: "^(aexp_to_str a))
    in res
  | Div (a0,a1) -> 
    let c0  = eval_aexp a0 sigma in
    let c1  = eval_aexp a1 sigma in
    let res = match (c0, c1) with
      | (IntConst(n1),IntConst(n2)) -> IntConst(n1/n2)
      | (RealConst(r1),RealConst(r2)) -> RealConst(r1/.r2)
      | (_,_) -> failwith ("Div aexp in eval_aexp is not well-typed: "^(aexp_to_str a))	
    in res
  | UnifReal(a0,a1) -> 
    let c0  = eval_aexp a0 sigma in
    let c1  = eval_aexp a1 sigma in
    let (lb, ub) = match (c0, c1) with
      | (IntConst(n1),IntConst(n2))   -> ((float_of_int n1),(float_of_int n2))
      | (IntConst(n1),RealConst(r2))  -> ((float_of_int n1),r2)
      | (RealConst(r1),IntConst(n2))  -> (r1,(float_of_int n2))
      | (RealConst(r1),RealConst(r2)) -> (r1,r2)
      | (_, _) -> failwith ("Invalid arguments to UnifReal in eval_aexp: "^(aexp_to_str a))
    in
      assert( lb < ub);
    let res = lb +. Random.float (ub-.lb) in
    RealConst(res)
  | UnifInt(a0,a1) -> 
    let c0  = eval_aexp a0 sigma in
    let c1  = eval_aexp a1 sigma in
    let (lb, ub) = match (c0, c1) with
      | (IntConst(n1),IntConst(n2)) -> (n1,n2)
      | (_,_) -> failwith ("Non-integer expressions in eval_aexp UnifInt: "^(aexp_to_str a))
    in
    let res = lb + Random.int (ub-lb) in
    IntConst(res)
  | _ -> failwith ("Unimplemented aexp in eval_aexp: "^(aexp_to_str a))

let rec eval_bexp (b: bexp) (sigma: concrete_state) : bool = match b with
  | True -> true
  | False -> false
  | Flip(p) ->
    let p' = eval_aexp p sigma in
    let p'' = 
      match p' with 
      | RealConst(r) -> r
      | _ -> failwith ("Incompatible expression inside Flip: "^(bexp_to_str b))
    in
    if ((Random.float 1.0) <= p'') then  true else false
  | EQ (a0, a1) -> 
    let a0' = eval_aexp a0 sigma in
    let a1' = eval_aexp a1 sigma in
    let res = match (a0', a1') with
      | (IntConst(n1),IntConst(n2)) -> n1 = n2
      | (RealConst(r1),RealConst(r2)) -> r1 = r2
      | (_,_) -> failwith ("EQ bexp in eval_bexp is not well-typed: "^(bexp_to_str b))
    in res
  | LE (a0, a1) -> 
    let a0' = eval_aexp a0 sigma in
    let a1' = eval_aexp a1 sigma in
    let res = match (a0', a1') with
      | (IntConst(n1),IntConst(n2)) -> n1 <= n2
      | (RealConst(r1),RealConst(r2)) -> r1 <= r2
      | (_,_) -> failwith ("LE bexp in eval_bexp is not well-typed: "^(bexp_to_str b))
    in res
  | LT (a0, a1) -> 
    let a0' = eval_aexp a0 sigma in
    let a1' = eval_aexp a1 sigma in
    let res = match (a0', a1') with
      | (IntConst(n1),IntConst(n2)) -> n1 < n2
      | (RealConst(r1),RealConst(r2)) -> r1 < r2
      | (_,_) -> failwith ("LT bexp in eval_bexp is not well-typed: "^(bexp_to_str b))
    in res
  | GE (a0, a1) -> 
    let a0' = eval_aexp a0 sigma in
    let a1' = eval_aexp a1 sigma in
    let res = match (a0', a1') with
      | (IntConst(n1),IntConst(n2)) -> n1 >= n2
      | (RealConst(r1),RealConst(r2)) -> r1 >= r2
      | (_,_) -> failwith ("GE bexp in eval_bexp is not well-typed: "^(bexp_to_str b))
    in res 
  | GT (a0, a1) -> 
    let a0' = eval_aexp a0 sigma in
    let a1' = eval_aexp a1 sigma in
    let res = match (a0', a1') with
      | (IntConst(n1),IntConst(n2)) -> n1 > n2
      | (RealConst(r1),RealConst(r2)) -> r1 > r2
      | (_,_) -> failwith ("GT bexp in eval_bexp is not well-typed: "^(bexp_to_str b))
    in res 
  | Not b -> 
    let b' = eval_bexp b sigma in
    (not b')
  | And (a, b) -> 
    let a' = eval_bexp a sigma in
    let b' = eval_bexp b sigma in
    (a' && b')
  | Or (a, b) -> 
    let a' = eval_bexp a sigma in
    let b' = eval_bexp b sigma in
    (a' || b')

let rec eval_com (c: com) (sigma: concrete_state) (path: int list) : (concrete_state * int list) = 
  match c with
    | Assign (Some(id), l, a) -> 
      let a' = eval_aexp a sigma in
      let sigma' = update sigma l a' in
      let path' = id :: path in
      (sigma', path')
    | If (Some(id), b, c0, None, _) ->
      let b' = eval_bexp b sigma in
      (* Record what branch was taken *)
      Util.branch_update id b'; 
      let path' = id :: path in
      let (sigma', path_modified) = 
	  if b' 
	  then(
	    let (sigma'', path'') = eval_com_list c0 sigma path' in
	    (sigma'', path''))
	  else 
	    (sigma, path')
      in
      let path_modified' = (id+path_prefix)::path_modified in
      (sigma', path_modified')
    | If (Some(id), b, c0, Some(c1), _) ->
      let b' = eval_bexp b sigma in
      (* Record what branch was taken *)
      Util.branch_update id b'; 
      let path' = id :: path in
      let (sigma', path_inner) = 
	if b' 
	then(
	  let (sigma'', path'') = eval_com_list c0 sigma path' in 
	  (sigma'', path''))
	else(
	  let (sigma'', path'') = eval_com_list c1 sigma path' in 
	  (sigma'', path''))
      in
      let path_inner' = (id+path_prefix)::path_inner in
      (sigma', path_inner')

    | While (Some(id), b, body, _) ->
      let b' = eval_bexp b sigma in
      (* Record whether body was taken *)
      Util.branch_update id b'; 
      let path' = id :: path in
      let (sigma'', path_inner) = 
	if b'
	then(
	  let (sigma', path'') = eval_com_list body sigma path' in
	  let (sigma'', path''') = eval_com (While(Some(id), b, body, None)) sigma' ((id+path_prefix)::path'') in
          (sigma'', path'''))
	else
	  (sigma, (id+path_prefix)::path')
      in (sigma'', path_inner)
    | Print (Some(id), a) -> 
      let a1 = eval_aexp a sigma in
      if !concrete_output then (print_endline (aexp_to_str a1));
      let path' = id :: path in
      (sigma, path')
    | Exp (_, a) -> failwith "Exp not implemented yet!"
    | Prob (Some(id), b) -> 
      let b' = eval_bexp b sigma in
      let path' = id :: path in
      if b' 
      then( 
	if !concrete_output then (print_endline "true");  (sigma, path'))
      else(
	if !concrete_output then (print_endline "false"); (sigma, path'))
    | Assume (Some(id), b) -> 
      let b' = eval_bexp b sigma in
      let path' = id :: path in
      if b' 
      then
	(sigma, path')
      else(
	failwith "eval_com: Assume violated!")
    | _ -> failwith ("Command not implemented or control location in command not set: "^(com_to_str c ""))

(* Think as call with empty path list and reverse append to current path at call site *)
and eval_com_list (com_list: com list) (sigma: concrete_state) (path_list: int list) : (concrete_state * int list) =
  List.fold_left (
    fun (agg_state,agg_p_list) elt -> eval_com elt agg_state agg_p_list
  ) (sigma, path_list) com_list

let rec eval_prog (c: com list) (sigma: concrete_state) (path: int list) : (concrete_state * int list) =
  let (sigma', path') = eval_com_list c sigma [] in
  (sigma', List.rev path')
