open Imp
open Util

let ssa_debug = ref false

type ssa_state = int LocMap.t
let empty_ssa_state: ssa_state = LocMap.empty

let ssa_lookup (ssa: ssa_state) (x: loc) : int =
  try
    LocMap.find x ssa
  with Not_found -> 0

let ssa_update (ssa: ssa_state) (x: loc) (n: int) : ssa_state = LocMap.add x n ssa

let rec ssa_aexp (a: aexp) (ssa_sigma: ssa_state) : aexp = match a with
  | IntConst _ -> a
  | RealConst _ -> a
  | Var x -> let ssa_num = ssa_lookup ssa_sigma x in
	     Var(x^"_"^(string_of_int ssa_num))
  | Add(a0, a1) ->
    let a0' = ssa_aexp a0 ssa_sigma in
    let a1' = ssa_aexp a1 ssa_sigma in
    Add(a0', a1')
  | Sub(a0, a1) ->
    let a0' = ssa_aexp a0 ssa_sigma in
    let a1' = ssa_aexp a1 ssa_sigma in
    Sub(a0', a1')
  | Mul(a0, a1) ->
    let a0' = ssa_aexp a0 ssa_sigma in
    let a1' = ssa_aexp a1 ssa_sigma in
    Mul(a0', a1')
  | Div(a0, a1) ->
    let a0' = ssa_aexp a0 ssa_sigma in
    let a1' = ssa_aexp a1 ssa_sigma in
    Div(a0', a1')
  | UnifReal(a0,a1) ->
    let a0' = ssa_aexp a0 ssa_sigma in
    let a1' = ssa_aexp a1 ssa_sigma in
    let (a0'', a1'') = (match (a0', a1') with
      | (IntConst(n1), IntConst(n2)) -> ((RealConst(float_of_int n1)),(RealConst(float_of_int n2)))
      | (IntConst(n1), RealConst(r2)) -> ((RealConst(float_of_int n1)), a1')
      | (RealConst(r1), IntConst(n2)) -> (a0',(RealConst(float_of_int n2)))
      | (RealConst(r1), RealConst(r2)) -> (a0', a1')
      | (_, _) -> failwith ("Invalid arguments to UnifReal in ssa_aexp: "^(aexp_to_str a)^". Arguments must be constants")
    ) in UnifReal(a0'', a1'')
  | UnifInt(a0,a1) ->
    let a0' = ssa_aexp a0 ssa_sigma in
    let a1' = ssa_aexp a1 ssa_sigma in
    let (a0'', a1'') = (match (a0', a1') with
      | (IntConst(n1), IntConst(n2)) -> (a0', a1')
      | _ -> failwith ("Non-integer expressions in ssa_aexp: "^(aexp_to_str a))
    ) in UnifInt(a0'', a1'')
  | _ -> failwith ("Unimplemented aexp in ssa_aexp: "^(aexp_to_str a))

let rec ssa_bexp (b: bexp) (ssa_sigma: ssa_state) : bexp = match b with
  | True -> True
  | False -> False
  | Flip(p) -> 
    let p' = ssa_aexp p ssa_sigma in
    Flip(p')
  | EQ(a0, a1) ->
    let a0' = ssa_aexp a0 ssa_sigma in
    let a1' = ssa_aexp a1 ssa_sigma in
    EQ(a0', a1')
  | LE(a0, a1) ->
    let a0' = ssa_aexp a0 ssa_sigma in
    let a1' = ssa_aexp a1 ssa_sigma in
    LE(a0', a1')
  | LT(a0, a1) ->
    let a0' = ssa_aexp a0 ssa_sigma in
    let a1' = ssa_aexp a1 ssa_sigma in
    LT(a0', a1')
  | GE(a0, a1) ->
    let a0' = ssa_aexp a0 ssa_sigma in
    let a1' = ssa_aexp a1 ssa_sigma in
    GE(a0', a1')
  | GT(a0, a1) ->
    let a0' = ssa_aexp a0 ssa_sigma in
    let a1' = ssa_aexp a1 ssa_sigma in
    GT(a0', a1')
  | Not(b_) -> 
    let b' = ssa_bexp b_ ssa_sigma in
    Not(b')
  | And(b0, b1) ->
    let b0' = ssa_bexp b0 ssa_sigma in
    let b1' = ssa_bexp b1 ssa_sigma in
    And(b0', b1')
  | Or(b0, b1) ->
    let b0' = ssa_bexp b0 ssa_sigma in
    let b1' = ssa_bexp b1 ssa_sigma in
    Or(b0', b1')
      
(* We use guided SSA along a feasible program path that was already realized, so no need for phi nodes *)
(* We do shallow, NON-RECURSIVE(!) exploration of AST nodes. This is enough as path is stored in ssa_comhash *)

let ssa_com (c: com) (ssa_sigma: ssa_state) : (com * ssa_state) = match c with
  | Assign(Some(id), l, a) ->
    let a' = ssa_aexp a ssa_sigma in
    let n = ssa_lookup ssa_sigma l in
    let ssa_sigma' = ssa_update ssa_sigma l (n+1) in
    let new_var_name = l^"_"^(string_of_int (n+1)) in
    (Assign(Some(id), new_var_name, a'), ssa_sigma')
  | If (Some(id), b, c0, c1, end_set) -> 
    let b' = ssa_bexp b ssa_sigma in
    (If(Some(id), b', c0, c1, end_set), ssa_sigma)
  | While (Some(id), b, bdy, end_set) ->
    let b' = ssa_bexp b ssa_sigma in
    (While(Some(id), b', bdy, end_set), ssa_sigma)
  | Print (Some(id), a) ->
    let a' = ssa_aexp a ssa_sigma in
    (Print(Some(id),a'), ssa_sigma)
  | Exp (Some(id), a) ->
    let a' = ssa_aexp a ssa_sigma in
    (Exp(Some(id),a'), ssa_sigma)
  | Prob(Some(id), b) ->
    let b' = ssa_bexp b ssa_sigma in
    (Prob(Some(id),b'), ssa_sigma)
  | Assume(Some(id), b) ->
    let b' = ssa_bexp b ssa_sigma in
    (Assume(Some(id), b'), ssa_sigma)
  | _ -> failwith ("ssa_com: "^(com_to_str c "")^" com not yet handled or annotation id not set!")

let ssa_path (path_list: int list) = (* returns a simple_com Simplecomhash *)
  let ssa_comhash = make_com_hash () in
  let (prog, _) = Preprocess.get_progs () in
  let path_prefix = ConcreteExec.path_prefix in

  let path_list_iterator = ref 0 in
  let _ = List.fold_left (
    fun agg_ssa elt_com_id ->
      if (elt_com_id < path_prefix) then
	let com = Comhash.find prog elt_com_id in
	let (com', ssa') = ssa_com com agg_ssa in
	Comhash.replace ssa_comhash !path_list_iterator (com');
	incr path_list_iterator;
	ssa'
      else
	agg_ssa
  ) empty_ssa_state path_list in

  if !ssa_debug then (
  let path_iter = ref 0 in
  List.iter (fun com_id -> 
    if com_id < path_prefix then (
      let updated_com = 
	try 
	  Comhash.find ssa_comhash !path_iter
	with Not_found -> failwith ("Ssacomhash command with id "^(string_of_int !path_iter)^" not found! "^"com_id = "^(string_of_int com_id))
      in
      print_endline ("["^(string_of_int com_id)^"]: "^(com_to_str updated_com ""));
      incr path_iter)
  ) path_list);
  ssa_comhash
