(*--- 

frontend: A prototype frontend for probabilistic programs

    Copyright (C) 2013  Aleksandar Chakarov <alch4231@colorado.edu>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

--*)

open Imp
open Util

let symbolic_debug = ref false

let r_index = ref 0

let next_r_index () : int = incr r_index; !r_index

let sym_lookup (sigma: symbolic_state) (x: loc) : aexp = 
  try
    LocMap.find x sigma
  with Not_found -> Var(x) 

let sym_update (sigma: symbolic_state) (x: loc) (n: aexp) : symbolic_state = LocMap.add x n sigma

let rec sym_eval_aexp (a: aexp) (sym_sigma: symbolic_state) : (aexp * symbolic_state) = match a with
  | IntConst _   -> (a, sym_sigma)
  | RealConst _  -> (a, sym_sigma)
  | Var x        -> let sym_res = sym_lookup sym_sigma x in
		    (sym_res, sym_sigma)
  | Add (a0, a1) -> let (a0', sym_sigma')  = sym_eval_aexp a0 sym_sigma  in
		    let (a1', sym_sigma'') = sym_eval_aexp a1 sym_sigma' in 
		    (Add(a0', a1'), sym_sigma'')
  | Sub (a0, a1) -> let (a0', sym_sigma')  = sym_eval_aexp a0 sym_sigma  in
		    let (a1', sym_sigma'') = sym_eval_aexp a1 sym_sigma' in 
		    (Sub(a0', a1'), sym_sigma'')
  | Mul (a0, a1) -> let (a0', sym_sigma')  = sym_eval_aexp a0 sym_sigma  in
		    let (a1', sym_sigma'') = sym_eval_aexp a1 sym_sigma' in 
		    (Mul(a0', a1'), sym_sigma'')
  | Div (a0, a1) -> let (a0', sym_sigma')  = sym_eval_aexp a0 sym_sigma  in
		    let (a1', sym_sigma'') = sym_eval_aexp a1 sym_sigma' in 
		    (Div(a0', a1'), sym_sigma'')
  | UnifReal (a0, a1) -> let (a0', sym_sigma')  = sym_eval_aexp a0 sym_sigma  in
			 let (a1', sym_sigma'') = sym_eval_aexp a1 sym_sigma' in 
			 let r_num = next_r_index () in
			 let var_name = "r_"^(string_of_int r_num) in
			 let sym_res  = UnifReal(a0', a1') in
			 let sym_sigma''' = sym_update sym_sigma'' var_name sym_res in
			 (Var(var_name), sym_sigma''')
  | UnifInt  (a0, a1) -> let (a0', sym_sigma')  = sym_eval_aexp a0 sym_sigma  in
			 let (a1', sym_sigma'') = sym_eval_aexp a1 sym_sigma' in 
			 let r_num = next_r_index () in
			 let var_name = "r_"^(string_of_int r_num) in
			 let sym_res  = UnifInt(a0', a1') in
			 let sym_sigma''' = sym_update sym_sigma'' var_name sym_res in
			 (Var(var_name), sym_sigma''')
  | _ -> failwith ("sym_eval_aexp: aexp "^(aexp_to_str a)^" not yet implemented!")

let rec sym_eval_bexp (b: bexp) (sym_sigma: symbolic_state) : (bexp * symbolic_state) = match b with
  | True  -> (True,  sym_sigma)
  | False -> (False, sym_sigma)
  | Flip(p) -> let (p', sym_sigma') = sym_eval_aexp p sym_sigma in
	       (Flip(p'), sym_sigma')
  | EQ (a0, a1) -> let (a0', sym_sigma' ) = sym_eval_aexp a0 sym_sigma  in
		   let (a1', sym_sigma'') = sym_eval_aexp a1 sym_sigma' in
		   (EQ(a0', a1'), sym_sigma'')
  | GE (a0, a1) -> let (a0', sym_sigma' ) = sym_eval_aexp a0 sym_sigma  in
		   let (a1', sym_sigma'') = sym_eval_aexp a1 sym_sigma' in
		   (GE(a0', a1'), sym_sigma'')
  | GT (a0, a1) -> let (a0', sym_sigma' ) = sym_eval_aexp a0 sym_sigma  in
		   let (a1', sym_sigma'') = sym_eval_aexp a1 sym_sigma' in
		   (GT(a0', a1'), sym_sigma'')
  | LE (a0, a1) -> let (a0', sym_sigma' ) = sym_eval_aexp a0 sym_sigma  in
		   let (a1', sym_sigma'') = sym_eval_aexp a1 sym_sigma' in
		   (LE(a0', a1'), sym_sigma'')
  | LT (a0, a1) -> let (a0', sym_sigma' ) = sym_eval_aexp a0 sym_sigma  in
		   let (a1', sym_sigma'') = sym_eval_aexp a1 sym_sigma' in
		   (LT(a0', a1'), sym_sigma'')
  | Not b_      -> let (b' , sym_sigma' ) = sym_eval_bexp b_ sym_sigma  in
		   (Not(b'), sym_sigma')
  | And(b0, b1) -> let (b0', sym_sigma' ) = sym_eval_bexp b0 sym_sigma  in
		   let (b1', sym_sigma'') = sym_eval_bexp b1 sym_sigma' in
		   (And(b0', b1'), sym_sigma'')
  | Or(b0, b1)  -> let (b0', sym_sigma' ) = sym_eval_bexp b0 sym_sigma  in
		   let (b1', sym_sigma'') = sym_eval_bexp b1 sym_sigma' in
		   (Or(b0', b1'), sym_sigma'')


let slice_ref = ref [-1]
let slice_iter = ref 0 
let get_current_slice () : int = 
  List.hd !slice_ref 
let advance_slice () : unit =
  slice_ref := List.tl !slice_ref

	     
(* Path formula is not be collected here. *)
let sym_eval_com (c: com) (sym_sigma: symbolic_state) (path_formula: bexp list) : (com * symbolic_state * bexp list) = match c with
  | Assign (Some(id), l, a)            -> let (a', sym_sigma') = sym_eval_aexp a sym_sigma in
					  let sym_sigma'' = sym_update sym_sigma' l a' in
					  if get_current_slice () == id then 
					    advance_slice (); 
				          (Assign(Some(id), l, a'), sym_sigma'', path_formula)
  | If (Some(id), b, c0, c1, end_set)  -> let next_flag = Util.get_next_branch_decision id in
					  (* let _ = print_endline (string_of_bool next_flag) in *)
					  let (b', sym_sigma') = sym_eval_bexp b sym_sigma in
					  let sym_constraint = if next_flag then b' else Not(b') in
					  let path_formula' = 
					    if get_current_slice () == id then (
					      advance_slice ();
					      sym_constraint :: path_formula)
					    else
					      path_formula
					  in
					  (If(Some(id), b', c0, c1, end_set), sym_sigma', path_formula')
  | While (Some(id), b, body, end_set) -> let next_flag = Util.get_next_branch_decision id in
					  (* let _ = print_endline (string_of_bool next_flag) in *)
					  let (b', sym_sigma') = sym_eval_bexp b sym_sigma in
					  let sym_constraint = if next_flag then b' else Not(b') in
					  let path_formula' = 
					    if get_current_slice () == id then (
					      advance_slice ();
					      sym_constraint :: path_formula)
					    else 
					      path_formula
					  in
					  (While(Some(id), b', body, end_set), sym_sigma', path_formula')
  | Print  (Some(id), a)               -> let (a', sym_sigma') = sym_eval_aexp a sym_sigma in
					  (Print(Some(id), a'), sym_sigma', path_formula)
  | Exp    (Some(id), a)               -> let (a', sym_sigma') = sym_eval_aexp a sym_sigma in
					  (Exp(Some(id), a'), sym_sigma', path_formula)
  | Prob   (Some(id), b)               -> let (b', sym_sigma') = sym_eval_bexp b sym_sigma in
					  (Prob(Some(id), b'), sym_sigma', b'::path_formula)
  | Assume (Some(id), b)               -> let (b', sym_sigma') = sym_eval_bexp b sym_sigma in
					  (Assume(Some(id), b'), sym_sigma', b'::path_formula)
  | _ -> failwith ("sym_eval_simple_com: simple_com "^(com_to_str c "")^" has no annotation id!")

let rec range_constructor (k: int) (range_nk: int list) : int list =
  if k > 0 then
    range_constructor (k-1) (k::range_nk)
  else
    0::range_nk
  
let range (n:int) : int list =
  range_constructor (n-1) []
    

let get_com_id (c: com) : int =
  match c with
    | Assign(Some(id), _, _) -> id
    | If(Some(id), _, _, _, _) -> id
    | While(Some(id), _, _, _) -> id
    | Print(Some(id), _) -> id
    | Exp(Some(id), _) -> id
    | Prob(Some(id), _) -> id
    | Assume(Some(id), _) -> id
    | _ -> failwith "guidedSymExec: next_com_id encountered a command that lacks ID"

let sym_eval_path (slice: int list) (ssa_comhash) = (* : (comhash * symbolic_state * bexp list * bexp) *)
  r_index := 0;
  slice_ref := slice;
  slice_iter := 0;
  let sym_comhash = make_com_hash () in
  let path_prefix = ConcreteExec.path_prefix in
  let path_list_iterator = ref 0 in
  let count = Comhash.length ssa_comhash in
  let range_count = range count in
 
  (*
  let iterating_fn ((sym_sigma, path_f): symbolic_state * bexp list) (elt_com_id: int) (next_elt_id: int) : symbolic_state * bexp list =
    if (elt_com_id < path_prefix) then (
      let com = Comhash.find ssa_comhash elt_com_id in
      let elt_id = get_com_id com in
      let next_id = if (next_elt_id >= 0) then (get_com_id (Comhash.find ssa_comhash next_elt_id)) else (-1) in
      let next_flag = if (elt_id == next_id - 1) then true else false in
      let (com', agg_sym_sigma', agg_pf') = sym_eval_com com next_flag sym_sigma path_f in
      if !symbolic_debug then(
	print_endline ("["^(string_of_int elt_id)^"]: "^(com_to_str com' "")));
      Comhash.replace sym_comhash !path_list_iterator com';
      incr path_list_iterator;
      (agg_sym_sigma', agg_pf'))
    else
      (sym_sigma, path_f)
  in
  let (final_sym_sigma, path_formula) = List.fold_left2 iterating_fn (empty_symbolic_state, []) range_count ((List.tl range_count)@[-1]) in *)
  
  let iterating_fn ((sym_sigma, path_f): symbolic_state * bexp list) (elt_com_id: int) : symbolic_state * bexp list =
    if (elt_com_id < path_prefix) then (
      let com = Comhash.find ssa_comhash elt_com_id in
      let elt_id = get_com_id com in
      let (com', agg_sym_sigma', agg_pf') = sym_eval_com com sym_sigma path_f in
      if !symbolic_debug then(
	print_endline ("["^(string_of_int elt_id)^"]: "^(com_to_str com' "")));
      Comhash.replace sym_comhash !path_list_iterator com';
      incr path_list_iterator;
      (agg_sym_sigma', agg_pf'))
    else
      (sym_sigma, path_f)
  in
  let (final_sym_sigma, path_formula) = List.fold_left iterating_fn (empty_symbolic_state, []) range_count in
  
  
  let property = List.hd path_formula in
  let path_formula' = List.rev (List.tl path_formula) in

  if !symbolic_debug then( 
    let formula_str_list = List.map bexp_to_str path_formula' in
    let formula_str = String.concat " /\\ " formula_str_list in
    
    print_endline "PATH FORMULA: ";
    print_endline formula_str;
    print_endline "PROPERTY FORMULA: ";
    print_endline (bexp_to_str property);
    print_endline ("BRANCHES SIZE: "^(string_of_int (Util.Branchhash.length !Util.branches)));

  );
  (final_sym_sigma, path_formula', property)
