type loc = string

module SSet = Set.Make(String)
type sset = SSet.t

type aexp = 
  | IntConst of int
  | RealConst of float
  | Var of loc                    (* x *)
  | Add of aexp * aexp            (* a0 + a1 *) 
  | Sub of aexp * aexp            (* a0 - a1 *) 
  | Mul of aexp * aexp            (* a0 * a1 *) 
  | Div of aexp * aexp            (* a0 / a1 *)
  | UnifReal of aexp * aexp       (* unifReal(0;1) *)
  | UnifInt of aexp * aexp        (* unifInt(0;10) *)
  | Gaussian of aexp * aexp       (* Gaussian(mean;stdev) *)
  | Laplacian of aexp * aexp      (* Laplacian(0;1) not sure about this yet*)

type bexp =
  | True
  | False
  | Flip of aexp                  (* flip(p)  *)
  | EQ of aexp * aexp             (* a0 == a1 *) 
  | LE of aexp * aexp             (* a0 <= a1 *) 
  | LT of aexp * aexp             (* a0 <  a1 *)
  | GE of aexp * aexp             (* a0 >= a1 *) 
  | GT of aexp * aexp             (* a0 >  a1 *)
  | Not of bexp                   (* !b *)
  | And of bexp * bexp            (* b0 /\ b1 *) (* b0 && b1 *) 
  | Or of bexp * bexp             (* b0 \/ b1 *) (* b0 || b1 *) 

type com =
  | Assign of int option * loc * aexp           (* @prog loc n: x := a *) 
  | If of int option * bexp * com list * (com list) option * sset option (* if_n b then c0 else c1 *)
  | While of int option * bexp * com list * sset option (* while_n (phi tuples) b do body *)
  | Print of int option * aexp                  (* print_n a *)
  | Exp of int option * aexp                    (* estimateExp_n(x) = Ex *)
  | Prob of int option * bexp                   (* estimateProb_n(x > y) *)
  | Assume of int option * bexp                 (* assume_n(b) *)

type prog = com list

type simple_com = 
  | SAssign of int option * loc * aexp
  | SAssume of int option * bexp
  | SPrint of int option * aexp
  | SExp of int option * aexp
  | SProb of int option * bexp

type simple_prog = simple_com list

(* Pretty-printing utility functions -- these aren't necessary but might be
 * handy for debugging. They convert instances of Aexp, Bexp and Com into
 * strings (that you can print out later with print_string, say). *) 
module P = Printf 

let rec aexp_to_str a = match a with
  | IntConst(n)     -> string_of_int n
  | RealConst(r)    -> string_of_float r
  | Var l -> l
  | Add (a, b)      -> let a_str = aexp_to_str a in
		       let b_str = aexp_to_str b in
		       "("^a_str^" + "^b_str^")"
  | Sub (a, b)      -> let a_str = aexp_to_str a in
		       let b_str = aexp_to_str b in
		       "("^a_str^" - "^b_str^")"
  | Mul (a, b)      -> let a_str = aexp_to_str a in
		       let b_str = aexp_to_str b in
		       "("^a_str^" * "^b_str^")"
  | Div (a, b)      -> let a_str = aexp_to_str a in
		       let b_str = aexp_to_str b in
		       "("^a_str^" / "^b_str^")"
  | UnifReal (a, b) -> let a_str = aexp_to_str a in
		       let b_str = aexp_to_str b in
		       "unifReal("^a_str^";"^b_str^")"
  | UnifInt  (a, b) -> let a_str = aexp_to_str a in
		       let b_str = aexp_to_str b in
		       "unifInt("^a_str^";"^b_str^")"
  | Gaussian (a, b) -> let a_str = aexp_to_str a in
		       let b_str = aexp_to_str b in
		       "gaussian("^a_str^";"^b_str^")"
  | Laplacian(a, b) -> let a_str = aexp_to_str a in
		       let b_str = aexp_to_str b in
		       "laplacian("^a_str^";"^b_str^")"

and bexp_to_str b = match b with
  | True       -> "true"
  | False      -> "false" 
  | Flip p     -> let p_str = aexp_to_str p in
		  "flip("^p_str^")"
  | EQ (a, b)  -> let a_str = aexp_to_str a in
		  let b_str = aexp_to_str b in
		  "("^a_str^" = "^b_str^")"
  | LE (a, b)  -> let a_str = aexp_to_str a in
		  let b_str = aexp_to_str b in
		  "("^a_str^" <= "^b_str^")"
  | LT (a, b)  -> let a_str = aexp_to_str a in
		  let b_str = aexp_to_str b in
		  "("^a_str^" < "^b_str^")" 
  | GE (a, b)  -> let a_str = aexp_to_str a in
		  let b_str = aexp_to_str b in
		  "("^a_str^" >= "^b_str^")" 
  | GT (a, b)  -> let a_str = aexp_to_str a in
		  let b_str = aexp_to_str b in
		  "("^a_str^" > "^b_str^")"
  | Not b      -> let b_str = bexp_to_str b in 
		  "!"^b_str
  | And (a, b) -> let a_str = bexp_to_str a in
		  let b_str = bexp_to_str b in
		  "("^a_str^" && "^b_str^")"
  | Or (a, b)  -> let a_str = bexp_to_str a in
		  let b_str = bexp_to_str b in
		  "("^a_str^" || "^b_str^")"

(* com_to_str takes a command c to convert to string and current indentation string *)
and com_to_str (c: com) (indent: string) : string = match c with
  | Assign (None, l, a)              -> let a_str = aexp_to_str a in
					indent^l^" = "^a_str
  | If (None, b, c0, None, _)        -> let b_str = bexp_to_str b in
					let extra_indent = indent^"\t" in
					let then_ = com_list_to_str c0 extra_indent in
					indent^"if "^b_str^"\n"^indent^"then\n"^
					  then_^indent^"end"  
  | If (None, b, c0, Some(c1), _)    -> let b_str = bexp_to_str b in
					let extra_indent = indent^"\t" in
					let then_ = com_list_to_str c0 extra_indent in
					let else_ = com_list_to_str c1 extra_indent in
					indent^"if "^b_str^"\n"^indent^"then\n"^
					  then_^indent^"else\n"^else_^indent^"end"
  | While (None, b, body, _)         -> let b_str = bexp_to_str b in
					let extra_indent = indent^"\t" in
					let body_ = com_list_to_str body extra_indent in
					indent^"while "^b_str^" do\n"^body_^indent^"end"
  | Print (None, a)                  -> let a_str = aexp_to_str a in
					indent^"print "^a_str
  | Exp (None, a)                    -> let a_str = aexp_to_str a in
					indent^"estimateExp "^a_str
  | Prob (None, b)                   -> let b_str = bexp_to_str b in
					indent^"estimateProb "^b_str
  | Assume (None, b)                 -> let b_str = bexp_to_str b in
					indent^"assume "^b_str
  | Assign (Some(n), l, a)           -> let n_str = string_of_int n in
					let a_str = aexp_to_str a in
					indent^"@loc "^n_str^": "^l^" = "^a_str
  | If (Some(n), b, c0, None, _) -> 
                                        let n_str = string_of_int n in
					let b_str = bexp_to_str b in
					let extra_indent = indent^"\t" in
					let then_ = com_list_to_str c0 extra_indent in
					indent^"if_"^n_str^" "^b_str^"\n"^indent^"then\n"^
					  then_^indent^"end"
  | If (Some(n), b, c0, Some(c1), _) -> 
                                        let n_str = string_of_int n in
				     	let b_str = bexp_to_str b in
					let extra_indent = indent^"\t" in
					let then_ = com_list_to_str c0 extra_indent in
					let else_ = com_list_to_str c1 extra_indent in
					indent^"if_"^n_str^" "^b_str^"\n"^indent^
					  "then\n"^then_^indent^"else\n"^else_^indent^"end"
  | While (Some(n), b, body,_) -> 
                                        let n_str = string_of_int n in
					let b_str = bexp_to_str b in
					let extra_indent = indent^"\t" in
					let body_ = com_list_to_str body extra_indent in
					indent^"while_"^n_str^" "^b_str^" do\n"^body_
					^indent^"end"
  | Print (Some(n), a)               -> let n_str = string_of_int n in
					let a_str = aexp_to_str a in
					indent^"print_"^n_str^" "^a_str
  | Exp (Some(n), a)                 -> let n_str = string_of_int n in
					let a_str = aexp_to_str a in
					indent^"estimateExp_"^n_str^" "^a_str
  | Prob (Some(n), b)                -> let n_str = string_of_int n in
					let b_str = bexp_to_str b in
					indent^"estimateProb_"^n_str^" "^b_str
  | Assume (Some(n), b)              -> let n_str = string_of_int n in
					let b_str = bexp_to_str b in
					indent^"assume_"^n_str^" "^b_str
  | _ -> failwith "com_to_str: Malformed command encountered"
(* A helper method to print blocks of commands *)
and com_list_to_str coms indent : string = 
  List.fold_left (
    fun agg elt -> let elt_str = com_to_str elt indent in
		   agg^elt_str^";\n"
  ) "" coms 
    
let prog_to_str p = com_list_to_str p "" 

let simple_com_to_str c = match c with
  | SAssign (None, l, a)           -> let a_str = aexp_to_str a in
				      l^" = "^a_str
  | SAssume (None, b)              -> let b_str = bexp_to_str b in
				      "assume "^b_str
  | SProb (None, b)                -> let b_str = bexp_to_str b in
				      "estimateProb "^b_str
  | SPrint (None, a)               -> let a_str = aexp_to_str a in
				      "print "^a_str
  | SExp (None, a)                 -> let a_str = aexp_to_str a in
				      "estimateExp "^a_str				

  | SAssign (Some(n), l, a)        -> let n_str = string_of_int n in
				      let a_str = aexp_to_str a in
				      "@loc "^n_str^": "^l^" = "^a_str
  | SAssume (Some(n), b)           -> let n_str = string_of_int n in
				      let b_str = bexp_to_str b in
				      "assume_"^n_str^" "^b_str
  | SPrint (Some(n), a)            -> let n_str = string_of_int n in
				      let a_str = aexp_to_str a in
				      "print_"^n_str^" "^a_str
  | SExp (Some(n), a)              -> let n_str = string_of_int n in
				      let a_str = aexp_to_str a in
				      "estimateExp_"^n_str^" "^a_str
  | SProb (Some(n), b)             -> let n_str = string_of_int n in
				      let b_str = bexp_to_str b in
				      "estimateProb_"^n_str^" "^b_str
					
let simple_prog_to_str p = 
  let prog_str_list = List.map simple_com_to_str p in
  String.concat ";\n" prog_str_list

