(*--- 

frontend: A prototype frontend for probabilistic programs

    Copyright (C) 2013  Aleksandar Chakarov <alch4231@colorado.edu>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

--*)

open Printf
open VolCompInterface



let unique_paths = Util.make_upath_hash ()

let main () =
  let started = ref 0.0 in
  started := Sys.time();
  let full_sigma = ref false in 
  let concrete = ref false in

  let args = [
    "--dir", Arg.Set_string  VolCompInterface.path_prefix, "Set path prefix: default /tmp";
    "--lim", Arg.Set_int PathSelection.limit_runs, "Set number of limit runs: default 90";
    "--concrete", Arg.Set concrete, "Show output of concrete execution";
    "--debug-path", Arg.Set PathSelection.path_debug, "Show path selection debug information";
    "--debug-slicing", Arg.Set PathSlice.slice_debug, "Show path slicing debug information";
    "--debug-ssa", Arg.Set GuidedSSA.ssa_debug, "Show SSA debug information";
    "--debug-sym-exec", Arg.Set GuidedSymExec.symbolic_debug, "Show Symbolic Execution debug information"
  ] in 
  Arg.parse args (fun _ -> ()) "" ; 
  
(* NEW CODE *)
  Random.self_init ();
  let lexbuf = Lexing.from_channel stdin in
  let imp_prog = try Parse.prog Lex.initial lexbuf with | Parsing.Parse_error -> (
    Printf.printf "Line number : %d \n" !Lex.lineNum ;
    raise (Failure "Parse Error")
  ) in
  let _ = print_endline "Original Program :" in
  let _ = print_endline (Imp.prog_to_str imp_prog) in
  let preproc_prog = Preprocess.preprocess_prog imp_prog in
  let _ = print_endline "\nPreprocessed Program :" in
  let _ = print_endline (Imp.prog_to_str preproc_prog) in
  let (prog, inverse_prog) = Preprocess.get_progs () in

  if !concrete then(print_endline "Concrete output flag is set. Concrete output follows:"; ConcreteExec.concrete_output := true);

  let unique_paths' = PathSelection.collect_paths preproc_prog unique_paths in
  let path_time = ref 0.0 in 
  path_time := Sys.time();
  let gen_time = !path_time -. !started in
  Util.Pathhash.iter (fun path (sym_sigma, path_f, r_index, props) ->
    let _ = encode_problem sym_sigma path_f r_index props in
    ()
  ) unique_paths';
  let write_time = ref 0.0 in 
  write_time := Sys.time();
  let proc_time = !write_time -. !path_time in
  let _ = prerr_endline ("Took: "^(string_of_float gen_time)^" to generate paths.") in
  let _ = prerr_endline ("Took: "^(string_of_float proc_time)^" to write paths out.") in
  let _ = prerr_endline ("Total Executions: "^(string_of_int !PathSelection.iteration)^".") in
  let _ = prerr_endline ("Unique Paths: "^(string_of_int !PathSelection.unique_paths_num)^".") in 

(* OLD CODE *)
(*
  let args = [
    ("--dir", Arg.Set_string  VolCompInterface.path_prefix, "Set path prefix: default /tmp");
    ("--lim", Arg.Set_int PathSelection.limit_runs, "Set number of limit runs: default 29");
    "--full", Arg.Set full_sigma, "Show full info";
    "--concrete", Arg.Set concrete, "Show output of concrete execution" 
  ] in 
  Arg.parse args (fun _ -> ()) "" ; 

  Random.self_init ();
  let lexbuf = Lexing.from_channel stdin in
  let imp_command =try Parse.com Lex.initial lexbuf with | Parsing.Parse_error -> (
    Printf.printf "Line number : %d \n" !Lex.lineNum ;
    raise (Failure "Parse Error")
  )in
  let _ = print_endline "Original Program :" in
  let _ = print_endline (Imp.com_to_str imp_command) in
  let preproc_command = Preprocess.preprocess_com imp_command in
  let _ = print_endline "\nPreprocessed Program :" in
  let _ = print_endline (Imp.com_to_str preproc_command) in
  let (prog, inverse_prog) = Preprocess.get_progs () in

  if !concrete then(print_endline "Concrete output flag is set. Concrete output follows:";SymExec.concrete_output := true);

  (*let c_state = SymExec.empty_concrete_state in
  let s_state = SymExec.empty_symbolic_state in
  let ssa     = SymExec.empty_ssa_state in
  let path_f  = [] in
  let path    = [] in
  let props   = [] in
  let (c_state', s_state', ssa', path_f', path', props') = 
    SymExec.eval_com preproc_command c_state s_state ssa path_f path props
  in
  let path_f_str = List.map Imp.bexp_to_str path_f' in
  let str_of_path_f = String.concat " && " path_f_str in
  let _ = print_endline ("Path Formula: ["^str_of_path_f^"]") in
  let path_str = List.map string_of_int path' in
  let str_of_path = String.concat ";" path_str in
  print_endline ("Path: ["^str_of_path^"]");*)
  let unique_paths' = PathSelection.collect_paths preproc_command unique_paths in
  let path_time = ref 0.0 in 
  path_time := Sys.time();
  let gen_time = !path_time -. !started in
  Util.Pathhash.iter (fun path (sym_sigma, path_f, r_index, props) ->
	let _ = encode_problem sym_sigma path_f r_index props in
    ()
  ) unique_paths';
  let write_time = ref 0.0 in 
  write_time := Sys.time();
  let proc_time = !write_time -. !path_time in
  let _ = prerr_endline ("Took: "^(string_of_float gen_time)^" to generate paths.") in
  let _ = prerr_endline ("Took: "^(string_of_float proc_time)^" to write paths out.") in
  let _ = prerr_endline ("Total Executions: "^(string_of_int !PathSelection.iteration)^".") in
  let _ = prerr_endline ("Unique Paths: "^(string_of_int !PathSelection.unique_paths_num)^".") in

*)  
  ()
  ;exit 0 
;;
main () ;;
