open Imp

let concrete_output = ref false

module LocMap = Map.Make(struct
  type t = loc
  let compare = compare
end)

type concrete_state = {real_state: (float LocMap.t); int_state: (int LocMap.t)}
type symbolic_state = aexp LocMap.t
type ssa_state = int LocMap.t

let empty_concrete_state: concrete_state = {real_state = LocMap.empty; int_state = LocMap.empty}
let empty_symbolic_state: symbolic_state = LocMap.empty
let empty_ssa_state: ssa_state = LocMap.empty

let r_index = ref 0

let lookup (sigma: concrete_state) (x: loc) : aexp = 
  let r_sigma = sigma.real_state in
  let i_sigma = sigma.int_state in
  let res = 
    try
      RealConst(LocMap.find x r_sigma)
    with Not_found -> (
      try
	IntConst(LocMap.find x i_sigma)
    with Not_found -> failwith ("Concrete lookup for var "^x^" failed."))
  in res

let update (sigma: concrete_state) (x: loc) (n: aexp) : concrete_state = 
  match n with 
    | IntConst(n') -> let i_sigma' = LocMap.add x n' sigma.int_state in
		      {real_state = sigma.real_state; int_state = i_sigma'}
    | RealConst(r') -> let r_sigma' = LocMap.add x r' sigma.real_state in
		       {real_state = r_sigma'; int_state = sigma.int_state}
    | _ -> failwith ("Unexpected expression for var "^x^" in concrete state : "^(aexp_to_str n))

let sym_lookup (sigma: symbolic_state) (x: loc) : aexp = 
  try
    LocMap.find x sigma
  with Not_found -> Var(x) 

let sym_update (sigma: symbolic_state) (x: loc) (n: aexp) : symbolic_state = LocMap.add x n sigma
let sym_sigma_to_str (sym_sigma: symbolic_state) : string = 
  let temp = (LocMap.fold (fun (v) (a) accum -> (accum^"("^v^" = "^(aexp_to_str a)^") , ")) sym_sigma "") in
  String.sub temp 0 ((String.length temp)-3)

let rec trans_to_str_help (sym_sigma: symbolic_state) (i: int) (n: int) (out: string) : string =
  if i >= (n+1)
  then
    out
  else
    let var_name = "r_"^(string_of_int i) in
    let out' = out^"("^var_name^" = "^(aexp_to_str (sym_lookup sym_sigma var_name))^") && " in
    (trans_to_str_help sym_sigma (i+1) n out')

let trans_to_str (sym_sigma: symbolic_state) : string =
  if !r_index > 0 
  then
    let temp = trans_to_str_help sym_sigma 1 !r_index "" in
    String.sub temp 0 ((String.length temp)-3)
  else
    "No random variables introduced!"

let ssa_lookup (ssa: ssa_state) (x: loc) : int =
  try
    LocMap.find x ssa
  with Not_found -> 0
let ssa_update (ssa: ssa_state) (x: loc) (n: int) : ssa_state = LocMap.add x n ssa

(* Concretely evaluate program, compute the symbolic expressions in almost ssa form *)
let rec eval_aexp (a: aexp) (sigma: concrete_state) (sym_sigma: symbolic_state) (ssa: ssa_state) : (aexp * aexp * symbolic_state) = match a with
  | IntConst n -> (IntConst n, IntConst(n), sym_sigma)
  | RealConst r -> (RealConst r, RealConst(r), sym_sigma)
  | Var x ->
    let res = lookup sigma x in
    let ssa_num = ssa_lookup ssa x in 
    let sym_res = sym_lookup sym_sigma (x^"_"^(string_of_int ssa_num)) in
    (res, sym_res, sym_sigma)
  | Add (a0,a1) -> 
    let (c0,aexp0,sym_sigma')  = eval_aexp a0 sigma sym_sigma  ssa in
    let (c1,aexp1,sym_sigma'') = eval_aexp a1 sigma sym_sigma' ssa in
    let res = match (c0, c1) with
      | (IntConst(n1),IntConst(n2)) -> IntConst(n1+n2)
      | (RealConst(r1),RealConst(r2)) -> RealConst(r1+.r2)
      | (_,_) -> failwith ("Add aexp in eval_aexp is not well-typed: "^(aexp_to_str a))
    in (res, Add(aexp0,aexp1), sym_sigma'') 
  | Sub (a0,a1) ->
    let (c0,aexp0,sym_sigma')  = eval_aexp a0 sigma sym_sigma  ssa in
    let (c1,aexp1,sym_sigma'') = eval_aexp a1 sigma sym_sigma' ssa in
    let res = match (c0, c1) with
      | (IntConst(n1),IntConst(n2)) -> IntConst(n1-n2)
      | (RealConst(r1),RealConst(r2)) -> RealConst(r1-.r2)
      | (_,_) -> failwith ("Sub aexp in eval_aexp is not well-typed: "^(aexp_to_str a))
    in (res, Sub(aexp0,aexp1), sym_sigma'') 
  | Mul (a0,a1) -> 
    let (c0,aexp0,sym_sigma')  = eval_aexp a0 sigma sym_sigma  ssa in
    let (c1,aexp1,sym_sigma'') = eval_aexp a1 sigma sym_sigma' ssa in
    let res = match (c0, c1) with
      | (IntConst(n1),IntConst(n2)) -> IntConst(n1*n2)
      | (RealConst(r1),RealConst(r2)) -> RealConst(r1*.r2)
      | (_,_) -> failwith ("Mul aexp in eval_aexp is not well-typed: "^(aexp_to_str a))
    in (res, Mul(aexp0,aexp1), sym_sigma'') 
  | Div (a0,a1) -> 
    let (c0,aexp0,sym_sigma')  = eval_aexp a0 sigma sym_sigma  ssa in
    let (c1,aexp1,sym_sigma'') = eval_aexp a1 sigma sym_sigma' ssa in
    let res = match (c0, c1) with
      | (IntConst(n1),IntConst(n2)) -> IntConst(n1/n2)
      | (RealConst(r1),RealConst(r2)) -> RealConst(r1/.r2)
      | (_,_) -> failwith ("Div aexp in eval_aexp is not well-typed: "^(aexp_to_str a))	
    in (res, Div(aexp0,aexp1), sym_sigma'') 
  | UnifReal(a0,a1) -> 
    let (c0,aexp0,sym_sigma')  = eval_aexp a0 sigma sym_sigma  ssa in
    let (c1,aexp1,sym_sigma'') = eval_aexp a1 sigma sym_sigma' ssa in
    let (lb, ub) = match (c0, c1) with
      | (IntConst(n1),IntConst(n2))   -> ((float_of_int n1),(float_of_int n2))
      | (IntConst(n1),RealConst(r2))  -> ((float_of_int n1),r2)
      | (RealConst(r1),IntConst(n2))  -> (r1,(float_of_int n2))
      | (RealConst(r1),RealConst(r2)) -> (r1,r2)
      | (_, _) -> failwith ("Invalid arguments to UnifReal in eval_aexp: "^(aexp_to_str a))
    in
      assert( lb < ub);
    let res = lb +. Random.float (ub-.lb) in
    r_index := !r_index + 1;
    let var_name = "r_"^(string_of_int !r_index) in 
    let sym_sigma' = sym_update sym_sigma var_name (UnifReal(RealConst(lb),RealConst(ub))) in
    let sym_res = Var(var_name) in
    (RealConst(res), sym_res, sym_sigma')
  | UnifInt(a0,a1) -> 
    let (c0,aexp0,sym_sigma')  = eval_aexp a0 sigma sym_sigma  ssa in
    let (c1,aexp1,sym_sigma'') = eval_aexp a1 sigma sym_sigma' ssa in
    let (lb, ub) = match (c0, c1) with
      | (IntConst(n1),IntConst(n2)) -> (n1,n2)
      | (_,_) -> failwith ("Non-integer expressions in eval_aexp UnifInt: "^(aexp_to_str a))
    in
    let res = lb + Random.int (ub-lb) in
    r_index := !r_index + 1;
    let var_name = "r_"^(string_of_int !r_index) in 
    let sym_sigma' = sym_update sym_sigma var_name (UnifInt(IntConst(lb),IntConst(ub))) in
    let sym_res = Var(var_name) in
    (IntConst(res), sym_res, sym_sigma')
  | _ -> failwith ("Unimplemented aexp in eval_aexp: "^(aexp_to_str a))

let rec eval_bexp (b: bexp) (sigma: concrete_state) (sym_sigma: symbolic_state) (ssa: ssa_state) : (bool * bexp * symbolic_state) = match b with
  | True -> (true, True, sym_sigma)
  | False -> (false, False, sym_sigma) 
  | Flip(p) ->
    let (p',aexpp,sym_sigma') = eval_aexp p sigma sym_sigma ssa in
    let p'' = 
      match p' with 
      | RealConst(r) -> r
      | _ -> failwith ("Incompatible expression inside Flip: "^(bexp_to_str b))
    in
    if ((Random.float 1.0) <= p'') then  (true, Flip(aexpp), sym_sigma') else (false, Flip(aexpp), sym_sigma')
  | EQ (a0, a1) -> 
    let (a0',aexp0,sym_sigma')  = eval_aexp a0 sigma sym_sigma  ssa in
    let (a1',aexp1,sym_sigma'') = eval_aexp a1 sigma sym_sigma' ssa in
    let res = match (a0', a1') with
      | (IntConst(n1),IntConst(n2)) -> n1 = n2
      | (RealConst(r1),RealConst(r2)) -> r1 = r2
      | (_,_) -> failwith ("EQ bexp in eval_bexp is not well-typed: "^(bexp_to_str b))
    in
    let sym_res = EQ(aexp0, aexp1) in
    (res, sym_res, sym_sigma'')
  | LE (a0, a1) -> 
    let (a0',aexp0,sym_sigma')  = eval_aexp a0 sigma sym_sigma  ssa in
    let (a1',aexp1,sym_sigma'') = eval_aexp a1 sigma sym_sigma' ssa in
    let res = match (a0', a1') with
      | (IntConst(n1),IntConst(n2)) -> n1 <= n2
      | (RealConst(r1),RealConst(r2)) -> r1 <= r2
      | (_,_) -> failwith ("LE bexp in eval_bexp is not well-typed: "^(bexp_to_str b))
    in
    let sym_res = LE(aexp0, aexp1) in
    (res, sym_res, sym_sigma'')
  | LT (a0, a1) -> 
    let (a0',aexp0,sym_sigma')  = eval_aexp a0 sigma sym_sigma  ssa in
    let (a1',aexp1,sym_sigma'') = eval_aexp a1 sigma sym_sigma' ssa in
    let res = match (a0', a1') with
      | (IntConst(n1),IntConst(n2)) -> n1 < n2
      | (RealConst(r1),RealConst(r2)) -> r1 < r2
      | (_,_) -> failwith ("LT bexp in eval_bexp is not well-typed: "^(bexp_to_str b))
    in
    let sym_res = LT(aexp0, aexp1) in
    (res, sym_res, sym_sigma'')
  | GE (a0, a1) -> 
    let (a0',aexp0,sym_sigma')  = eval_aexp a0 sigma sym_sigma  ssa in
    let (a1',aexp1,sym_sigma'') = eval_aexp a1 sigma sym_sigma' ssa in
    let res = match (a0', a1') with
      | (IntConst(n1),IntConst(n2)) -> n1 >= n2
      | (RealConst(r1),RealConst(r2)) -> r1 >= r2
      | (_,_) -> failwith ("GE bexp in eval_bexp is not well-typed: "^(bexp_to_str b))
    in
    let sym_res = GE(aexp0, aexp1) in
    (res, sym_res, sym_sigma'')
  | GT (a0, a1) -> 
    let (a0',aexp0,sym_sigma')  = eval_aexp a0 sigma sym_sigma  ssa in
    let (a1',aexp1,sym_sigma'') = eval_aexp a1 sigma sym_sigma' ssa in
    let res = match (a0', a1') with
      | (IntConst(n1),IntConst(n2)) -> n1 > n2
      | (RealConst(r1),RealConst(r2)) -> r1 > r2
      | (_,_) -> failwith ("GT bexp in eval_bexp is not well-typed: "^(bexp_to_str b))
    in
    let sym_res = GT(aexp0, aexp1) in
    (res, sym_res, sym_sigma'')  
  | Not b -> 
    let (b',bexpb,sym_sigma')  = eval_bexp b sigma sym_sigma  ssa in
    (not b', Not(bexpb), sym_sigma')
  | And (a, b) -> 
    let (a',bexpa,sym_sigma')  = eval_bexp a sigma sym_sigma  ssa in
    let (b',bexpb,sym_sigma'') = eval_bexp b sigma sym_sigma' ssa in
    (a' && b', And(bexpa,bexpb), sym_sigma'')
  | Or (a, b) -> 
    let (a',bexpa,sym_sigma')  = eval_bexp a sigma sym_sigma  ssa in
    let (b',bexpb,sym_sigma'') = eval_bexp b sigma sym_sigma' ssa in
    (a' || b', Or(bexpa,bexpb), sym_sigma'')

let rec eval_com (c: com) (sigma: concrete_state)(sym_sigma: symbolic_state) (ssa: ssa_state) (path_formula: bexp list) (path: int list) (props: bexp list) : (concrete_state * symbolic_state * ssa_state * bexp list * int list * bexp list) = 
  match c with
    | Assign (Some(id), l, a) -> 
      let (a',aexpa,sym_sigma') = eval_aexp a sigma sym_sigma ssa in
      let sigma' = update sigma l a' in
      let n = ssa_lookup ssa l in
      let ssa' = ssa_update ssa l (n+1) in
      let var_name = l^"_"^(string_of_int (n+1)) in
      let sym_sigma'' = sym_update sym_sigma' var_name aexpa in
      let path' = path @[id] in
      (sigma', sym_sigma'', ssa', path_formula, path', props)
    | If (Some(id), b, c0, None) ->
      let (b', bexpb, sym_sigma') = eval_bexp b sigma sym_sigma ssa in
      let (sigma', sym_sigma'', ssa', path_formula'', path'', props') = 
	  if b' 
	  then(
	    let path_formula' = path_formula @ [bexpb] in
	    let path'         = path @ [id] in
	    eval_com c0 sigma sym_sigma' ssa path_formula' path' props)
	  else 
	    let path_formula' = path_formula @ [Not(bexpb)] in
	    let path'         = path @ [id] in
	    (sigma, sym_sigma', ssa, path_formula', path', props)
      in
      (sigma', sym_sigma'', ssa', path_formula'', path'', props')
    | If (Some(id), b_, c0, Some(c1)) ->
      let (b',bexpb,sym_sigma') = eval_bexp b_ sigma sym_sigma ssa in
      let (sigma', sym_sigma'', ssa', path_formula', path', props') = 
	if b' 
	then(
	  let path_formula'' = path_formula @ [bexpb] in
	  let path''         = path @ [id] in 
	  eval_com c0 sigma sym_sigma' ssa path_formula'' path'' props) 
	else(
	  let path_formula'' = path_formula @ [Not(bexpb)] in
	  let path''         = path @ [id] in
	  eval_com c1 sigma sym_sigma' ssa path_formula'' path'' props)
      in
      (sigma', sym_sigma'', ssa', path_formula', path', props')

    | While (Some(id), b_, c) ->
      let (b', bexpb, sym_sigma') = eval_bexp b_ sigma sym_sigma ssa in
      if (b')
      then(
	let path_formula'' = path_formula @ [bexpb] in
	let path''         = path @ [id] in
	let (sigma', sym_sigma'', ssa', path_formula', path', props') = eval_com c sigma sym_sigma' ssa path_formula'' path'' props in
        (eval_com (While(Some(id),b_, c)) sigma' sym_sigma'' ssa' path_formula' path' props'))
      else(
	let path_formula'' = path_formula @ [Not(bexpb)] in
	let path''         = path @ [id] in
	(sigma, sym_sigma', ssa, path_formula'', path'', props))
    | Print (Some(id), a) -> 
      let (a1, _, sym_sigma') = eval_aexp a sigma sym_sigma ssa in
      if !concrete_output then (print_endline (aexp_to_str a1));
      let path''         = path @ [id] in
      (sigma, sym_sigma', ssa, path_formula, path'', props)
    | Exp (_, a) -> failwith "Exp not implemented yet!"
    | Prob (Some(id), b) -> 
      let (b', bexpb, sym_sigma') = eval_bexp b sigma sym_sigma ssa in
      let path''         = path @ [id] in
      let props' = props @ [bexpb] in
      if b' 
      then( 
	if !concrete_output then (print_endline "true");  (sigma, sym_sigma', ssa, path_formula, path'', props'))
      else(
	if !concrete_output then (print_endline "false"); (sigma, sym_sigma', ssa, path_formula, path'', props'))
    | Assume (Some(id), b) -> 
      let (b', bexpb, sym_sigma') = eval_bexp b sigma sym_sigma ssa in
      let path''         = path @ [id] in
      if b' 
      then
	(sigma, sym_sigma', ssa, path_formula, path, props) (* And(path, bexpb))) *)
      else
	(sigma, sym_sigma', ssa, path_formula, path, props) (* And(path, bexpb))) *)
    | _ -> failwith ("Command not implemented or control location in command not set: "^(com_to_str c))

let rec eval_prog (c: com) (sigma: concrete_state) (sym_sigma: symbolic_state) (ssa: ssa_state) (path_f: bexp list) (path: int list) (props: bexp list) : (concrete_state * symbolic_state * ssa_state * bexp list * int list * bexp list) =
  r_index := 0;
  let target_props = [] in
  let (sigma', sym_sigma', ssa', path_f', path', target_props') = eval_com c sigma sym_sigma ssa path_f path target_props in
  (sigma', sym_sigma', ssa', path_f', path', target_props')
