(*--- 

frontend: A prototype frontend for probabilistic programs

    Copyright (C) 2013  Aleksandar Chakarov <alch4231@colorado.edu>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

--*)

open Imp

let slice_debug = ref false
let empty_set: sset = SSet.empty

let print_live_set (live: sset) (label: string) = 
  let live_set_elts = SSet.elements live in
  let live_set_str  = String.concat ", " live_set_elts in 
  print_endline (label^": {"^live_set_str^"}")

let rec live_set_aexp (a: aexp) (live: sset) : sset = match a with
  | IntConst _ -> live
  | RealConst _ -> live
  | Var x -> SSet.add x live
  | Add(a0, a1) -> let live0 = live_set_aexp a0 live  in
		   let live1 = live_set_aexp a1 live0 in
		   live1
  | Sub(a0, a1) -> let live0 = live_set_aexp a0 live  in
		   let live1 = live_set_aexp a1 live0 in
		   live1
  | Mul(a0, a1) -> let live0 = live_set_aexp a0 live  in
		   let live1 = live_set_aexp a1 live0 in
		   live1
  | Div(a0, a1) -> let live0 = live_set_aexp a0 live  in
		   let live1 = live_set_aexp a1 live0 in
		   live1
  | UnifReal(a0, a1) -> let live0 = live_set_aexp a0 live  in
			let live1 = live_set_aexp a1 live0 in
			live1
  | UnifInt(a0, a1) -> let live0 = live_set_aexp a0 live  in
		       let live1 = live_set_aexp a1 live0 in
		       live1
  | _ -> failwith ("live_set_aexp: aexp "^(aexp_to_str a)^" not implemented")

let rec live_set_bexp (b: bexp) (live: sset) : sset = match b with
  | True        -> live
  | False       -> live
  | Flip(p)     -> let livep =  live_set_aexp p live in
		   livep
  | EQ(a0, a1)  -> let live0 = live_set_aexp a0 live in
		   let live1 = live_set_aexp a1 live0 in
		   live1
  | LE(a0, a1)  -> let live0 = live_set_aexp a0 live in
		   let live1 = live_set_aexp a1 live0 in
		   live1
  | LT(a0, a1)  -> let live0 = live_set_aexp a0 live in
		   let live1 = live_set_aexp a1 live0 in
		   live1
  | GE(a0, a1)  -> let live0 = live_set_aexp a0 live in
		   let live1 = live_set_aexp a1 live0 in
		   live1
  | GT(a0, a1)  -> let live0 = live_set_aexp a0 live in
		   let live1 = live_set_aexp a1 live0 in
		   live1
  | Not b_      -> let liveb = live_set_bexp b_ live in
		   liveb
  | And(b0, b1) -> let live0 = live_set_bexp b0 live in
		   let live1 = live_set_bexp b1 live0 in
		   live1
  | Or(b0, b1)  -> let live0 = live_set_bexp b0 live in
		   let live1 = live_set_bexp b1 live0 in
		   live1

(* This simple approach keeps all branch guards and assignaments to live vars in these *)
let live_set_simple (c: com) (live: sset) (kept_path: int list) : (sset * int list) = match c with
  | Assign(Some(id), l, a) ->
    if (SSet.mem l live) then
      let live' = SSet.remove l live in
      let lvsa  = live_set_aexp a empty_set in
      let liveU = SSet.union live' lvsa in 
      (liveU, id::kept_path)
    else (* If assigning to non-live var, don't change live set and don't add to path *)
      (live, kept_path) 

  | If(Some(id), b, c0, None, _) ->
    let lvsb  = live_set_bexp b empty_set in
    let liveU = SSet.union live lvsb in
    (liveU, id::kept_path)
      
  | If(Some(id), b, c0, Some(c1), _) ->
    let lvsb  = live_set_bexp b empty_set in
    let liveU = SSet.union live lvsb in
    (liveU, id::kept_path)
      
  | While(Some(id), b, body, _) ->
    let lvsb  = live_set_bexp b empty_set in
    let liveU = SSet.union live lvsb in
    (liveU, id::kept_path)
      
  | Assume(Some(id), b) ->
    let lvsb  = live_set_bexp b empty_set in
    let liveU = SSet.union live lvsb in
    (liveU, id::kept_path) 
  
  | _ -> (* Print, Exp, Prob *)
    (live, kept_path)
    
let rec rec_write_set_com (c: com) (ws: sset) : sset =
  match c with
    | Assign (Some(id), l, a) ->
      SSet.add l ws
    | If (Some(id), b, c0, None, end_set) ->
      let wsT = rec_write_set_com_list c0 empty_set in
      SSet.union ws wsT
    | If (Some(id), b, c0, Some(c1), end_set) ->
      let wsT = rec_write_set_com_list c0 empty_set in
      let wsE = rec_write_set_com_list c1 empty_set in
      SSet.union ws (SSet.union wsT wsE)
    | While (Some(id), b, body, end_set) ->
      let wsBdy = rec_write_set_com_list body empty_set in
      SSet.union ws wsBdy
    | _ -> ws

and rec_write_set_com_list (com_list: com list) (ws: sset) : sset =
  List.fold_left (fun agg_ws elt -> rec_write_set_com elt agg_ws) ws com_list


let rec live_set_com (c: com) (live: sset) (ws: sset) (kept_path: int list) : (sset * sset * int list * com) = match c with
  | Assign(Some(id), l, a) ->
    (* We don't need assignments in the symbolic execution phase *)
    let (liveU, kpU) = 
      if (SSet.mem l live) then (
	let live' = SSet.remove l live in
	let lvsa  = live_set_aexp a empty_set in
	if (!slice_debug) then(
	  print_endline ("KEEP ASSIGN to "^l^" @ "^(string_of_int id))
	);
	(* Remove assignments *)
	(SSet.union live' lvsa, id::kept_path))
	(*(SSet.union live' lvsa, kept_path)) *)
      else(
	if (!slice_debug) then(
	  print_endline ("SKIP ASSIGN to "^l^" @ "^(string_of_int id));
	);
	(live, kept_path))
    in
    let ws'   = SSet.add l ws in
    (liveU, ws', kpU, c)
      
  | If(Some(id), b, c0, None, end_set) ->
    let lvsb  = live_set_bexp b empty_set in
    let liveU = SSet.union live lvsb in
    (match end_set with
      | None -> (live, ws, kept_path, If(Some(id), b, c0, None, Some(live)))
      | Some(es) -> 
	let wsT = rec_write_set_com_list c0 empty_set in
	if (!slice_debug) then(
	  print_live_set es "END SET";
	  print_live_set wsT "WST SET";
	);
	if (SSet.is_empty (SSet.inter es wsT))
	then(
	  if (!slice_debug) then(
	    print_endline ("SKIP IF-THEN @ "^(string_of_int id)));
	  assert(SSet.equal live es);
	  (live, ws, kept_path, If(Some(id), b, c0, None, None)))
	else(
	  if (!slice_debug) then(
	    print_endline ("KEEP IF-THEN @ "^(string_of_int id)));
	  (liveU, ws, id::kept_path, If(Some(id), b, c0, None, None)))
    )
 
  | If(Some(id), b, c0, Some(c1), end_set) ->
    let lvsb  = live_set_bexp b empty_set in
    let liveU = SSet.union live lvsb in
    (match end_set with
      | None -> (live, ws, kept_path, If(Some(id), b, c0, Some(c1), Some(live)))
      | Some(es) ->
	let wsT =  rec_write_set_com_list (List.rev c0) empty_set in
	let wsE =  rec_write_set_com_list (List.rev c1) empty_set in
	let ws_both_branches = SSet.union wsT wsE in
	if (!slice_debug) then(
	  print_live_set es "END SET";
	  print_live_set wsT "WST SET";
	  print_live_set wsE "WSE SET";
	  print_live_set ws_both_branches "WST_U_WSE SET");
	if (SSet.is_empty (SSet.inter ws_both_branches es))
	then(
	  if (!slice_debug) then(
	    print_endline ("SKIP IF-THEN-ELSE @ "^(string_of_int id)));
	  assert(SSet.equal live es);
	  (live, ws, kept_path, If(Some(id), b, c0, Some(c1), None)))
	else(
	  if (!slice_debug) then(
	    print_endline ("KEEP IF-THEN-ELSE @ "^(string_of_int id)));
	  (liveU, ws, id::kept_path, If(Some(id), b, c0, Some(c1), None)))
    )
      
  | While(Some(id), b, body, end_set) ->
    let lvsb  = live_set_bexp b empty_set in
    let liveU = SSet.union live lvsb in
    (match end_set with
      | None -> (live, ws, kept_path, While(Some(id), b, body, Some(live)))
      | Some(es) ->
	let wsBdy = rec_write_set_com_list (List.rev body) empty_set in
	if (!slice_debug) then(
	  print_live_set es "END SET";
	  print_live_set wsBdy "WSBODY SET");
	if (SSet.is_empty (SSet.inter es wsBdy)) 
	then(
	  if (!slice_debug) then(
	    print_endline ("SKIP WHILE @ "^(string_of_int id)));
	  assert(SSet.equal live es);
	  (live, ws, kept_path, While(Some(id), b, body, None)))
	else(
	  if (!slice_debug) then(
	    print_endline ("KEEP WHILE @ "^(string_of_int id)));
	  (liveU, ws, id::(kept_path), While(Some(id), b, body, None)))
    )
      
  | Assume(Some(id), b) ->
    let lvsb  = live_set_bexp b empty_set in
    let liveU = SSet.union live lvsb in
    (liveU, ws, id::kept_path, c) 
  
  | _ -> (* Print, Exp, Prob *)
    (live, ws, kept_path, c)
      
and live_set_com_list (com_list: com list) (live_set: sset) (wr: sset) (kept_path: int list) : (sset * sset * int list) = 
  List.fold_left 
    (fun (agg_live_set, agg_wr, agg_kp) elt -> 
      let (ls, w, kp, _) = live_set_com elt agg_live_set agg_wr agg_kp in
      (ls, w, kp)
    ) (live_set, wr, kept_path) com_list
  
				    
let get_property (c: com) : bexp =
  match c with 
    | Prob (Some(id), b) -> b
    | _ -> failwith ("PathSlice: get_property: Property "^(com_to_str c "")^" not handled.")

(* Path Slice collects the relevant symbolc constraints wrt property along the path path_list *)
let path_slice (path_list: int list) (comhash) : int list =
  let path = List.rev path_list in
  let rev_path = List.rev path_list in
  let prop_id = List.hd rev_path in 
  let path_prefix = ConcreteExec.path_prefix in 
  let property_com = Util.Comhash.find comhash (if (prop_id > path_prefix) then (prop_id - path_prefix) else prop_id) in
  let property = get_property property_com in
  let live_set  = live_set_bexp property empty_set in 
  
  (* The Simple Approach that keeps all assignments to Live vars and all guards
  let (final_live_set, final_kept) = List.fold_left (
    fun (agg_live_set, agg_kept) elt ->
      let agg_set_elts = SSet.elements agg_live_set in
      let agg_set_str  = String.concat ", " agg_set_elts in 
      let _ = print_endline ("LIVE SET: {"^agg_set_str^"}") in
      if (elt < path_prefix) then  
	let elt_id = elt in
	let com = Util.Comhash.find comhash elt_id in
	let _ = print_endline ("["^(string_of_int elt)^"]: "^(com_to_str com "")) in
	let (ls, kp) = live_set_simple com agg_live_set agg_kept in
	(ls, kp)
      else (* If it is a join node *)
	let _ = print_endline ("["^(string_of_int elt)^"]: SKIPPED (JOIN NODE)") in
	(agg_live_set, agg_kept)
  ) (live_set, []) (List.tl rev_path) in *)

  (* The Approach that keeps assignments to Live vars and considers dropping guards if not relevant *)
  let (final_live_set, final_ws, final_kept) = List.fold_left (
    fun (agg_live_set, agg_ws, agg_kept) elt ->
      let elt_id = if (elt < path_prefix) then  elt else (elt - path_prefix) in
      let com = Util.Comhash.find comhash elt_id in
      if (!slice_debug) then(
	print_live_set agg_live_set "LIVE SET";
	print_endline ("["^(string_of_int elt)^"]: "^(com_to_str com "")));
      let (ls, ws, kp, com') = live_set_com com agg_live_set agg_ws agg_kept in
      let comhash = Util.Comhash.replace comhash elt_id com' in
      (ls, ws, kp)
  ) (live_set, empty_set, []) (List.tl rev_path) in
  if (!slice_debug) then(
    print_live_set final_live_set "PROPERTY SET");
  final_kept@[prop_id]
