(*--- 

frontend: A prototype frontend for probabilistic programs

    Copyright (C) 2013  Aleksandar Chakarov <alch4231@colorado.edu>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

--*)

open Imp
let loc_num = ref 0

let program = Util.make_com_hash()
let inv_program = Util.make_inverse_com_hash()

let preprocess_aexp (a: aexp) : aexp = 
  a  

let preprocess_bexp (b: bexp) : bexp =
  b

(* Add location information to every command *)
let rec preprocess_com (c: com) : (com * int) = match c with
  | Assign (n, l, a) -> (
    match n with
      | Some (n') -> failwith ("Preprocess.preprocess_com: Assign already has location "^(string_of_int n')^" assigned.")
      | None -> (
	let id = !loc_num in
	loc_num := !loc_num + 1;
	let c' = Assign(Some(id), l, a) in
	let _  = Util.Comhash.replace program id c' in
	let _  = Util.InverseComhash.replace inv_program c' id in
	(c', id))
  )
  | If (n, b, c0, None, None) ->(
    match n with
      | Some (n') -> failwith ("Preprocess.preprocess_com: Assign already has location "^(string_of_int n')^" assigned.")
      | None -> (
	let id = !loc_num in
	loc_num := !loc_num + 1;
	let (c0', t_end_id) = preprocess_com_list c0 in
	let c' = If(Some(id), b, c0', None, None) in
	let _  = Util.Comhash.replace program id c' in
	let _  = Util.InverseComhash.replace inv_program c' id in
	(c', id))
  )
  | If (n, b, c0, Some(c1), None)    -> (
    match n with
      | Some (n') -> failwith ("Preprocess.preprocess_com: Assign already has location "^(string_of_int n')^" assigned.")
      | None -> (
	let id = !loc_num in
	loc_num := !loc_num + 1;
	let (c0', t_end_id) = preprocess_com_list c0 in
	let (c1', e_end_id) = preprocess_com_list c1 in
	let c' = If(Some(id), b, c0', Some(c1'), None) in
	let _  = Util.Comhash.replace program id c' in
	let _  = Util.InverseComhash.replace inv_program c' id in
	(c', id))
  )
  | While (n, b, c0, None) ->(
    match n with
      | Some (n') -> failwith ("Preprocess.preprocess_com: Assign already has location "^(string_of_int n')^" assigned.")
      | None -> (
	let id = !loc_num in
	loc_num := !loc_num + 1;
	let (c0', bdy_end_id) = preprocess_com_list c0 in
	let c' = While(Some(id), b, c0', None) in
	let _  = Util.Comhash.replace program id c' in
	let _  = Util.InverseComhash.replace inv_program c' id in
	(c', id))
  )
  | Print (n, a) ->(
    match n with
      | Some (n') -> failwith ("Preprocess.preprocess_com: Assign already has location "^(string_of_int n')^" assigned.")
      | None -> (
	let id = !loc_num in
	loc_num := !loc_num + 1;
	let c' = Print(Some(id), a) in
	let _  = Util.Comhash.replace program id c' in
	let _  = Util.InverseComhash.replace inv_program c' id in
	(c', id))
  )
  | Exp (n, a) ->(
    match n with
      | Some (n') -> failwith ("Preprocess.preprocess_com: Assign already has location "^(string_of_int n')^" assigned.")
      | None -> (
	let id = !loc_num in
	loc_num := !loc_num + 1;
	let c' = Exp(Some(id), a) in
	let _  = Util.Comhash.replace program id c' in
	let _  = Util.InverseComhash.replace inv_program c' id in
	(c', id))
  )
  | Prob (n, b) ->(
    match n with
      | Some (n') -> failwith ("Preprocess.preprocess_com: Assign already has location "^(string_of_int n')^" assigned.")
      | None -> (
	let id = !loc_num in
	loc_num := !loc_num + 1;
	let c' = Prob(Some(id), b) in 
	let _  = Util.Comhash.replace program id c' in
	let _  = Util.InverseComhash.replace inv_program c' id in
	(c', id))
  )
  | Assume (n, b) ->(
    match n with
      | Some (n') -> failwith ("Preprocess.preprocess_com: Assign already has location "^(string_of_int n')^" assigned.")
      | None -> (
	let id = !loc_num in
	loc_num := !loc_num + 1;
	let c' = Assume(Some(id), b) in
	let _  = Util.Comhash.replace program id c' in
	let _  = Util.InverseComhash.replace inv_program c' id in
	(c', id))
  )

and preprocess_com_list (com_list: com list) : (com list * int) =
  let (ret_list, ret_id) = List.fold_left (
    fun (agg_list,agg_id) elt -> 
      let (elt_', id') = preprocess_com elt in
      (elt_'::agg_list, id')
  ) ([], 0) com_list in
  (List.rev ret_list, ret_id)

let preprocess_prog (p : com list) : com list = 
  let (pp_prog, _) = preprocess_com_list p in
  pp_prog

let get_progs () = 
  (program, inv_program)
