open Imp
open Unix

let path_prefix = ref "/tmp"
let r_max_index = ref 0
let s_sigma = ref Util.empty_symbolic_state

(* Do the actual encodings for every variable according to the distribution it is drawn from. See example below. *)

let rec build_struct_constraints_helper (sym_sigma: Util.symbolic_state) (curr_index: int) (r_index: int) (acc: string) = 
  if curr_index >= (r_index+1)
  then
    acc
  else
    let var_name = "r_"^(string_of_int curr_index) in
    let sym_res = GuidedSymExec.sym_lookup sym_sigma var_name in
    let (dist, param1, param2) = 
      match sym_res with
	| UnifReal (RealConst(r0),RealConst(r1)) -> 
	  ("U", (string_of_float r0), (string_of_float r1))
	| UnifInt  (IntConst(n0),IntConst(n1)) ->
	  ("I", (string_of_int n0), (string_of_int n1))
	| Gaussian (RealConst(r0),RealConst(r1)) ->
	  ("G", (string_of_float r0), (string_of_float r1))
	| Laplacian (RealConst(r0),RealConst(r1)) ->
	  ("L", (string_of_float r0), (string_of_float r1))
	| _ -> failwith ("Structural Constraints: Distribution for var "^var_name^" has aexp "^(aexp_to_str sym_res)^" not well defined.")
    in
    let acc' = acc^"V "^var_name^" "^dist^" "^param1^" "^param2^"\n" in
    (build_struct_constraints_helper sym_sigma (curr_index+1) r_index acc')


(* Structural constraints are of the form: V r_1 U 0.9 1.1, where V signifies variable, r_1 is var name, 
   U is underlying distribution, and 0.9, 1.1 are parameters to the distribution *)

let build_struct_constraints (sym_sigma: Util.symbolic_state) (r_index: int) =
  build_struct_constraints_helper sym_sigma 1 r_index ""


(* var_index returns the index i-1 in the array that corresponds to variable r_i *)

let var_index (x: Imp.loc) = 
  let index = String.index x '_' in
  let var_suffix = String.sub x (index+1) ((String.length x)-(index+1)) in
  let id = (int_of_string var_suffix)-1 in
  id


(* Macro to add to constants of the same type *)

let add_to_const (c: aexp) (amt: aexp) = 
  match (c, amt) with 
    | (IntConst(c'),IntConst(amt')) -> IntConst(c'+amt')
    | (RealConst(c'),RealConst(amt')) -> RealConst(c'+.amt')
    | _ -> failwith "Aexp not well typed in encode_aexp : add_to_const"


(* Creates an array of coefficients of the corresponding type for r_-variables. Each r_i variable's coefficient is zero (0 or 0.0) initially *)
let make_default_array () = 
  let arr = Array.make !r_max_index (RealConst(0.0)) in
  let _ = Array.iteri (fun i _ ->
			 let var_name = "r_"^(string_of_int (i+1)) in
			 let res = (match (GuidedSymExec.sym_lookup !s_sigma var_name) with
				      | UnifInt (_,_) -> IntConst(0)
				      | UnifReal (_,_) -> RealConst(0.0)
				      | Gaussian (_,_) -> RealConst(0.0)
				      | Laplacian (_,_) -> RealConst(0.0)
				      | a -> failwith ("Unexpected r_i aexp encountered in make_default_array: "^(aexp_to_str a))
				   ) in
			   Array.set arr i res) arr
  in arr


let rec getConstant (a:aexp) = 
  match a with 
    | IntConst n -> true, (float_of_int n)
    | RealConst r -> true, r
    | Var x0 -> false, 0.
    | Add (a0, a1) -> 
	let (t1,c1)  = getConstant a0 in 
	let (t2,c2) = getConstant a1 in 
	  (t1 && t2), (c1 +. c2)
    | Sub (a0, a1) -> 
	let (t1,c1)  = getConstant a0 in 
	let (t2,c2) = getConstant a1 in 
	  (t1 && t2), (c1 -. c2)
    | Mul (a0, a1) -> 
	let (t1,c1)  = getConstant a0 in 
	let (t2,c2) = getConstant a1 in 
	  (t1 && t2), (c1 *. c2)
    | _ -> failwith "unexpected expression"

let rec isIntType (a:aexp) = 
  match a with 
    | IntConst n -> true
    | RealConst r -> false
    | Var x0 ->
	let i = var_index x0 in 
	let var_name = "r_"^(string_of_int (i+1)) in
	  (match (GuidedSymExec.sym_lookup !s_sigma var_name) with
	     | UnifInt (_,_) -> true
	     | UnifReal (_,_) -> false
	     | Gaussian (_,_) -> false
	     | Laplacian (_,_) -> false
	     | a -> failwith ("Unexpected r_i aexp encountered in make_default_array: "^(aexp_to_str a))
	  )
    | Sub (a0, a1) 
    | Mul (a0, a1)  
    | Add (a0, a1) -> (
	let t0 = isIntType a0 in 
	let t1 = isIntType a1 in
	  match (t0, t1) with 
	    | true , true -> true
	    | true, false
	    | false, true -> failwith "type mismatch"
	    | false, false -> false
      )
    | _ -> failwith "unexpected expression"


   
let rec encode_aexp_new (a:aexp) arr scale = match a with 
  | IntConst n -> 
      let n' = (float_of_int n) in 
	(scale *. n'), arr 
  | RealConst r ->  (scale *. r), arr
  | Var x0 -> 
      let i = var_index x0 in 
	arr.(i) <- arr.(i) +. scale;
	(0.0,  arr)
  | Add ( a0 , a1) ->
      let (c0, arr0) = encode_aexp_new a0 arr scale  in 
      let (c1, arr1) = encode_aexp_new a1 arr0 scale in
	(c0 +. c1), arr1
  | Sub (a0, a1) -> 
      let (c0, arr0) = encode_aexp_new a0 arr scale  in 
      let (c1, arr1) = encode_aexp_new a1 arr0 (-1.0 *. scale) in
	(c0 +. c1), arr1
  | Mul (a0, a1) -> begin
      let (t0, c0) = getConstant a0 in 
      let (t1, c1) = getConstant a1 in 
	match (t0, t1) with 
	  | false, false -> failwith "nonlinear expression encountered"
	  | true, false -> encode_aexp_new a1 arr (scale *. c0) 
	  | false, true -> encode_aexp_new a0 arr (scale *. c1) 
	  | true, true -> ((c0 *. c1) /. scale), arr
    end
  | _ -> failwith "unexpected subexpression" 

let printConstraint (c,arr) = 
  let out_str = ref "C" in 
    for i = 0 to ((!r_max_index)-1)
    do
      out_str := (!out_str)^" "^(string_of_float arr.(i))
    done;
    out_str := (!out_str)^" "^(string_of_float c);
    out_str := (!out_str)^"\n";
    !out_str



let rec encode_bexp_new (b:bexp) = 
  let arr = Array.make !r_max_index 0.0 in 
  match b with 
    | LE (a0, a1) -> 
	begin
	  let (c0, arr0) = encode_aexp_new a0 arr 1.0 in 
	  let (c1, arr1) = encode_aexp_new a1 arr0 (-1.0) in
	    printConstraint ( ((-1.0) *. (c0 +. c1)), arr1 ) 
	end
    | LT (a0, a1) -> 
	begin
	  let (c0, arr0) = encode_aexp_new a0 arr 1.0 in 
	  let (c1, arr1) = encode_aexp_new a1 arr0 (-1.0) in
	    if (isIntType a0) 
	    then (
	      assert (isIntType a1);
	      printConstraint ( (  (-1.0) *. (c0 +. c1 +. 1.0)) , arr1)
	    ) else (
	      printConstraint ( ( (-1.0) *. (c0 +. c1)), arr1) 
	    )
	end 
    | GE (a0, a1) -> 
	encode_bexp_new (LE (a1, a0))
    | GT (a0, a1) -> 
	encode_bexp_new (LT (a1, a0))
    | Not b1 -> 
	let b' = (match b1 with
		    | LE (a0, a1) -> GT (a0,a1)
		    | LT (a0, a1) -> GE (a0,a1)
		    | GE (a0, a1) -> LT (a0,a1)
		    | GT (a0, a1) -> LE (a0,a1)
		    | Flip p ->  assert false
		    | _ -> failwith ("Something went wrong in VolCompInterface.encode_bexp: "^(bexp_to_str b))
		 ) in
	encode_bexp_new b'
    | And (b1,b2) -> (encode_bexp_new b1); (encode_bexp_new b2)
    | Or _ -> failwith "please avoid using || or && operators in if statements. We support && for probabilistic queries."
    | _ -> failwith " unexpected subexpression bexp"

(* Takes care of linearizing the aexp. Combined with encode_bexp, we get expressions in linear form *)

let rec encode_aexp (a: aexp) (arr) = match a with
  | IntConst(n) -> (a, arr)
  | RealConst(r) -> (a, arr)
  | Var (x0) -> 
      let type0 = Array.get arr (var_index x0) in
      let (c, index0) = (match type0 with
			   | IntConst(n0) -> (IntConst(0), IntConst(n0+1))
			   | RealConst(r0) -> (RealConst(0.0), RealConst(r0+.1.0))
			   | _ -> failwith ("Add expression is not well formed in encode_aexp: "^(aexp_to_str a))
			) in
      let _ = Array.set arr (var_index x0) index0 in
	(c, arr)
  | Add (a0,a1) -> 
    let (arr0, arr1) = (make_default_array (), make_default_array ()) in
    let (c0, arr0') = encode_aexp a0 arr0 in
    let (c1, arr1') = encode_aexp a1 arr1 in
    let new_c = (match (c0,c1) with
      | (IntConst(n0),IntConst(n1)) -> IntConst(n0+n1)
      | (RealConst(r0),RealConst(r1)) -> RealConst(r0+.r1)
      | (_,_) -> failwith ("Unexpected Add aexp in encode_aexp: "^(aexp_to_str a))
    )in
    let new_arr = make_default_array () in
    let _ = Array.iteri 
      (fun i c0 ->
	let c1 = Array.get arr1' i in
	let c' = add_to_const c0 c1 in
	Array.set new_arr i c'
      ) arr0'
    in
    (new_c, new_arr)
  | Sub (a0,a1) -> 
    let (arr0, arr1) = (make_default_array (), make_default_array ()) in
    let (c0, arr0') = encode_aexp a0 arr0 in
    let (c1, arr1') = encode_aexp a1 arr1 in
    let new_c = (match (c0,c1) with
      | (IntConst(n0),IntConst(n1)) -> IntConst(n0-n1)
      | (RealConst(r0),RealConst(r1)) -> RealConst(r0-.r1)
      | (_,_) -> failwith ("Unexpected Add aexp in encode_aexp: "^(aexp_to_str a))
    )in
    let new_arr = make_default_array () in
    let _ = Array.iteri 
      (fun i c0 ->
	let c1 = Array.get arr1' i in
	let c' = (match (c0,c1) with
	  | (IntConst(n0),IntConst(n1)) -> IntConst(n0-n1)
	  | (RealConst(r0),RealConst(r1)) -> RealConst(r0-.r1)
	  | (_,_) -> failwith ("Unexpected Add aexp in encode_aexp: "^(aexp_to_str a))
	) in
	Array.set new_arr i c'
      ) arr0'
    in
    (new_c, new_arr)
  | Mul (a0,a1) -> 
    (* NOTE: This is intended only for linear expressions. It will not work in a non-linear case, it will drop terms! *)
    let (arr0, arr1) = (make_default_array (), make_default_array ()) in
    let new_arr = make_default_array () in
    let (c0, arr0') = encode_aexp a0 arr0 in
    let check0 = (abs (compare arr0' new_arr)) > 0 in (* array is different than all-zeros? *)
    let (c1, arr1') = encode_aexp a1 arr1 in
    let check1 = (abs (compare arr1' new_arr)) > 0 in (* array is different than all-zeros? *)
    let new_c = (match (c0,c1) with
		   | (IntConst(n0),IntConst(n1)) -> IntConst(n0*n1)
      | (RealConst(r0),RealConst(r1)) -> RealConst(r0*.r1)
      | (_,_) -> failwith ("Unexpected Add aexp in encode_aexp: "^(aexp_to_str a))
    )in
    
    let (n,a) = (match (check0, check1) with
      | (false, false) -> (new_c, new_arr)
      | (false, true) -> 
	let _ = Array.iteri 
	  (fun i e0 ->
	    let e' = (match (e0,c1) with
	      | (IntConst(n0),IntConst(n1)) -> IntConst(n1*n0)
	      | (RealConst(r0),RealConst(r1)) -> RealConst(r0*.r1)
	      | (_,_) -> e0 (* MAKE SURE YOU TYPE CHECKED, OTHERWISE THIS IS WRONG!!! *)
	    ) in
	    Array.set new_arr i e'
	  ) arr0
	in
	(new_c,new_arr)
      | (true, false) ->
	let _ = Array.iteri 
	  (fun i e1 ->
	    let e' = (match (c0,e1) with
	      | (IntConst(n0),IntConst(n1)) -> IntConst(n0*n1)
	      | (RealConst(r0),RealConst(r1)) -> RealConst(r0*.r1)
	      | (_,_) -> e1 (* MAKE SURE YOU TYPE CHECKED, OTHERWISE THIS IS WRONG!!! *)
	    ) in
	    Array.set new_arr i e'
	  ) arr1
	in
	(new_c,new_arr)
      | (true, true) -> failwith ("We support only linear expressions. Non-linearity detected in aexp: "^(aexp_to_str a))
    ) in
    (n, a)
      
  | _ -> failwith ("Unimplemented aexp in encode_aexp: "^(aexp_to_str a))
    

let rec encode_bexp (b: bexp) = match b with
  | LE (a0, a1) -> 
    let (arr0, arr1) = (make_default_array (), make_default_array ()) in
    let (c0, arr0') = encode_aexp a0 arr0 in
    let (c1, arr1') = encode_aexp a1 arr1 in
    let new_c = (match (c0, c1) with
      | (IntConst(n0),IntConst(n1)) -> (string_of_int (n1-n0))
      | (RealConst(r0), RealConst(r1)) -> (string_of_float (r1-.r0))
      | _ -> failwith ("Type error in bexp: "^(bexp_to_str b))
    ) in
    let out_str = ref "" in
    out_str := "C";
    let _ = Array.iteri (
      fun i e0 ->
	let e1 = Array.get arr1' i in
	match (e0, e1) with
	  | (IntConst(n0), IntConst(n1)) -> out_str := !out_str^" "^(string_of_int (n0-n1))
	  | (RealConst(r0), RealConst(r1)) -> out_str := !out_str^" "^(string_of_float (r0-.r1))
	  | (_, _) -> failwith ("Type error in encode_bexp at bexp: "^(bexp_to_str b))
    ) arr0' in
    (!out_str^" "^new_c^"\n")

  | LT (a0, a1) -> 
    let (arr0, arr1) = (make_default_array (), make_default_array ()) in
    let (c0, arr0') = encode_aexp a0 arr0 in
    let (c1, arr1') = encode_aexp a1 arr1 in
    let new_c = (match (c0, c1) with
      | (IntConst(n0),IntConst(n1)) -> (string_of_int (n1-n0-1))
      | (RealConst(r0), RealConst(r1)) -> (string_of_float (r1-.r0))
      | _ -> failwith ("Type error in bexp: "^(bexp_to_str b))
    ) in
    let out_str = ref "" in
    out_str := "C";
    let _ = Array.iteri (
      fun i e0 ->
	let e1 = Array.get arr1' i in
	match (e0, e1) with
	  | (IntConst(n0), IntConst(n1)) -> out_str := !out_str^" "^(string_of_int (n0-n1))
	  | (RealConst(r0), RealConst(r1)) -> out_str := !out_str^" "^(string_of_float (r0-.r1))
	  | (_, _) -> failwith ("Type error in encode_bexp at bexp: "^(bexp_to_str b))
    ) arr0' in
    (!out_str^" "^new_c^"\n")
  | GE (a0, a1) -> 
    let (arr0, arr1) = (make_default_array (), make_default_array ()) in
    let (c0, arr0') = encode_aexp a0 arr0 in
    let (c1, arr1') = encode_aexp a1 arr1 in
    let new_c = (match (c0, c1) with
      | (IntConst(n0),IntConst(n1)) -> (string_of_int (n0-n1))
      | (RealConst(r0), RealConst(r1)) -> (string_of_float (r0-.r1))
      | _ -> failwith ("Type error in bexp: "^(bexp_to_str b))
    ) in
    let out_str = ref "" in
    out_str := "C";
    let _ = Array.iteri (
      fun i e0 ->
	let e1 = Array.get arr1' i in
	match (e0, e1) with
	  | (IntConst(n0), IntConst(n1)) -> out_str := !out_str^" "^(string_of_int (n1-n0))
	  | (RealConst(r0), RealConst(r1)) -> out_str := !out_str^" "^(string_of_float (r1-.r0))
	  | (_, _) -> failwith ("Type error in encode_bexp at bexp: "^(bexp_to_str b))
    ) arr0' in
    (!out_str^" "^new_c^"\n")
  | GT (a0, a1) -> 
    let (arr0, arr1) = (make_default_array (), make_default_array ()) in
    let (c0, arr0') = encode_aexp a0 arr0 in
    let (c1, arr1') = encode_aexp a1 arr1 in
    let new_c = (match (c0, c1) with
      | (IntConst(n0),IntConst(n1)) -> (string_of_int (n0-n1-1))
      | (RealConst(r0), RealConst(r1)) -> (string_of_float (r0-.r1))
      | _ -> failwith ("Type error in bexp: "^(bexp_to_str b))
    ) in
    let out_str = ref "" in
    out_str := "C";
    let _ = Array.iteri (
      fun i e0 ->
	let e1 = Array.get arr1' i in
	match (e0, e1) with
	  | (IntConst(n0), IntConst(n1)) -> out_str := !out_str^" "^(string_of_int (n1-n0))
	  | (RealConst(r0), RealConst(r1)) -> out_str := !out_str^" "^(string_of_float (r1-.r0))
	  | (_, _) -> failwith ("Type error in encode_bexp at bexp: "^(bexp_to_str b))
    ) arr0' in
    (!out_str^" "^new_c^"\n")
  | Not b_ -> (* MAKE SURE THERE ARE NO DOUBLE NEGATIONS FIRST *)
    let b' = (match b_ with
      | LE (a0, a1) -> GT (a0,a1)
      | LT (a0, a1) -> GE (a0,a1)
      | GE (a0, a1) -> LT (a0,a1)
      | GT (a0, a1) -> LE (a0,a1)
      | Flip p -> b_
      | _ -> failwith ("Something went wrong in VolCompInterface.encode_bexp: "^(bexp_to_str b))
    ) in
    encode_bexp b'
  | And (a0, a1) -> failwith "There should be no And bexp in encode_bexp."
  | Or (a0, a1)  -> failwith "There should be no Or bexp in encode_bexp."
  | _ -> ""




let file_num = ref 0

let encode_problem (sym_sigma: Util.symbolic_state) (path_f: bexp list) (r_index: int) (props: bexp list) = 
  r_max_index := r_index;
  s_sigma   := sym_sigma;
  (*let path_start = ref 0.0 in
  let proc_end = ref 0.0 in
  let write_end = ref 0.0 in
  
  path_start := Sys.time();*)
  let struct_constr_str = build_struct_constraints sym_sigma r_index in
 
  let path_f = List.map encode_bexp_new path_f in
  let str_of_path_f = String.concat "" path_f in

  let props = List.map encode_bexp_new props in
  (*proc_end := Sys.time();*)

  (* Print the Vars and Path constraints without any properties *)
  let file = open_out (!path_prefix^"/"^(string_of_int !PathSelection.limit_runs)^"/test-"^(string_of_int !file_num)^".in") in
  let _ = output_string file struct_constr_str in
  let _ = output_string file str_of_path_f in
  let _ = output_string file "E" in
  let _ = close_out file in


  
  (* Print Vars, Path constraints and one Property (phi-i) per file *)
  let phi_num = ref 0 in
  List.iter (
    fun prop_str ->
      let file = open_out (!path_prefix^"/"^(string_of_int !PathSelection.limit_runs)^"/test-"^(string_of_int !file_num)^"-phi-"^(string_of_int !phi_num)^".in") in
      let _ = output_string file struct_constr_str in
      let _ = output_string file str_of_path_f in
      let _ = output_string file prop_str in
      let _ = output_string file "E" in
      let _ = close_out file in
      phi_num := !phi_num + 1;
  ) props;
  (*write_end := Sys.time();*)
(*    let _ =  prerr_endline ("Wrote path #: "^(string_of_int !file_num)^" to file.") in *)
    file_num := !file_num + 1
    
