/*---

volComp: A prototype Volume Computation Tool.

    Copyright (C) 2013  Sriram Sankaranarayanan <srirams@colorado.edu>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

---*/
#ifndef D__BOX__HH__
#define D__BOX__HH__

#include "myUtils.hh"
#include "rVar.hh"

extern vector<RVar> varData;
class Box {
private:
  int nVars;
  vector<bool> hasLB;
  vector<long double> lb;
  vector<bool> hasUB;
  vector<long double> ub;
  

public:

  Box():
    nVars((int) varData.size()), 
    hasLB(varData.size(), false),
    hasUB(varData.size(), false),
    lb(varData.size()),
    ub(varData.size()){
    int i;
    for (i= 0; i < nVars; ++i){
      if (varData[i].hasLowerBound()){
	hasLB[i] = true;
	lb[i]  = varData[i].getLowerBound();
      } else {
	hasLB[i] = false;
      }
      if (varData[i].hasUpperBound()){
	hasUB[i] = true;
	ub[i] = varData[i].getUpperBound();
      } else{
	hasUB[i]= false;
      }
       
    }
  }
  void roundDimension(int dim)  {
    if (hasLB[dim]){
      lb[dim] = ceil(lb[dim]);
    }
    if (hasUB[dim]){
      ub[dim] = floor(ub[dim]);
    }
  }

  bool checkBounds(int dim) const {
    if (hasLB[dim] && hasUB[dim]) {
      assert(lb[dim] <= ub[dim]);
      if (lb[dim] <= ub[dim]) return false;
    }
    return true;
  }
  bool  hasLowerBound(int dim) const {
    assert(dim >= 0);
    assert(dim < nVars);
    return hasLB[dim];
  }
  
  bool hasUpperBound(int dim) const {
    assert(dim >= 0);
    assert(dim < nVars);
    return hasUB[dim];
  }

  bool hasBothBounds(int dim) const {
    assert(dim >= 0);
    assert(dim < nVars);
    return hasUB[dim] && hasLB[dim];
  }

  long double lowerBound(int dim) const {
    assert(dim >= 0);
    assert(dim < nVars);
    assert(hasLB[dim]);
    return lb[dim];
  }
  
  long double upperBound(int dim) const {
    assert(dim >= 0);
    assert(dim < nVars);
    assert(hasUB[dim]);
    return ub[dim];
  }

  void setUpperBound(int dim, long double v){
    assert(dim >= 0);
    assert(dim < nVars);
    if (varData[dim].isIntegerVariable())
      v = floor(v);
    if (!hasUB[dim] || ub[dim] >= v){
      hasUB[dim]=true;
      ub[dim]=v;
    }
    checkBounds(dim);
  }

  void  setLowerBound(int dim, long double v){
    assert(dim >= 0);
    assert(dim < nVars);
    if (varData[dim].isIntegerVariable())
      v = ceil(v);
    if (!hasLB[dim] || lb[dim] <= v) {
      hasLB[dim]=true;
      lb[dim]=v;
    }
    checkBounds(dim);
  }

  void unsetLowerBound(int dim){
    assert(dim >= 0);
    assert(dim < nVars);
    hasLB[dim]=false;
  }

  void unsetUpperBound(int dim){
    assert(dim >= 0);
    assert(dim < nVars);
    hasUB[dim]=false;
  }

  friend ostream & operator << (ostream & out, Box const & bx){
    int i;
    for (i = 0; i < bx.nVars; ++i){
      if (bx.hasLowerBound(i)){
	out << bx.lowerBound(i) << " <= ";
      } else {
	out << "-oo <= ";
      }
      out << varData[i].getName() << " <= ";
      if (bx.hasUpperBound(i)){
	out << bx.upperBound(i) << endl;
      } else {
	out << "oo" << endl;
      }

    }
    return out;
  }

};

#endif
