/*---

volComp: A prototype Volume Computation Tool.

    Copyright (C) 2013  Sriram Sankaranarayanan <srirams@colorado.edu>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

---*/
#include <vector> 
#include <ostream>
#include <map>
#include "myUtils.hh"
#include "rVar.hh"
#include "box.hh"
#include "unionFind.hh"

extern vector<RVar> varData;

class Ineq {
protected:
  /* -- 
     Represents the constraint 
         coeffs . x <= c
     ---*/
  std::vector<long double> coeffs;
  long double c;
  bool isRedundant;
public:
  Ineq(int nVars ):coeffs(nVars,0.0), 
		   c(-1.0),
		   isRedundant(false){}

  bool isBoxConstraint(int & dim, bool & isLB, long double & l, 
		       bool & isUB, long double & u) const {
    if (isRedundant) return false;
    int count = 0;
    int i;
    isLB=isUB=false;
    int nVars = (int) coeffs.size();
    for (i = 0; i < nVars; ++i){
      if (coeffs[i] > 0.0){
	if (count == 0){
	  dim = i;
	  assert (!isLB && !isUB);
	  isUB=true;
	  u = c/coeffs[i];
	  count = 1;
	} else {
	  return false;
	}
	count ++;
      } else if (coeffs[i] < 0.0) {
	if (count == 0){
	  dim = i;
	  assert (!isLB && !isUB);
	  isLB=true;
	  l = c/coeffs[i];
	} else {
	  return false;
	}
	count ++;
      }
    }
    return true;
  }

  long double & operator[] (int i){
    PRE( i >= 0);
    PRE( i < (int) coeffs.size());
    return coeffs[i];
  }

  long double operator() (int i) const {
    PRE( i >= 0);
    PRE( i < (int) coeffs.size());
    return coeffs[i];
  }
  
  long double const & constCoeff() const{
    return c;
  }

  long double & constCoeff(){
    return c;
  }

  void setConstant(long double x){
    assert(!isRedundant);
    c = x;
  }

  bool isSatisfied( vector<long double> const & p) const {
    if (isRedundant) return true;
    PRE (p.size() == coeffs.size());
    int i;
    long double r = 0.0;
    for (i = 0; i < p.size(); ++i){
      r += coeffs[i] * p[i];
    }
    return (r <= c);
  }

  void collectVarOccurrenceCounts( map<int,int> & oMap) const{
    if (isRedundant) return ;
    int i;
    for (i = 0; i < coeffs.size(); ++i){
      if (coeffs[i] > EPS || coeffs[i] < -EPS) {
	oMap[i] = oMap[i] +1;
      }
    }
  }

  int nVars() const {
    if (isRedundant) return 0;
    int i;
    int count = 0;
    for (i = 0; i < coeffs.size(); ++i){
      if (coeffs[i] > EPS || coeffs[i] < -EPS) {
	count ++;
      }
    }
    return count;
  }


  bool boxSatisfies(Box const & bx) const {
    int i;
    if (isRedundant)
      return true;

    long double b=0.0;
    // Compute upper bound on LHS and see if it is still less than the constant coeff.
    for (i = 0; i < coeffs.size(); ++i){
      if (coeffs[i] < 0.0){
	//Use lower bound
	if (!bx.hasLowerBound(i)) return false; // No way!
	b = b + bx.lowerBound(i) * coeffs[i];
      } 
      if (coeffs[i] > 0.0){
	// use upper bound
	if (!bx.hasUpperBound(i)) return false; // Again, no way!!
	b = b + bx.upperBound(i) * coeffs[i];
      }
    }

    return (b <= c);
  }

  bool subsumesOrIsSubsumed(Ineq const & in2, bool &subsumes, bool & isSubsumed, long double & scale) const {
    // Check if this subsumes in2 or vice-versa.
    scale = 0.0;
    int i = 0;
    subsumes=isSubsumed=false;
    for (i = 0; i < coeffs.size() && coeffs[i] == 0.0 && in2.coeffs[i] == 0.0; ++i){}
    if ( i < coeffs.size()){
      if (coeffs[i] == 0.0 || in2.coeffs[i] == 0.0) return false;
      scale = coeffs[i]/in2.coeffs[i];
      if (scale < 0.0) return false;
      i++;
      for (; i < coeffs.size(); ++i){
	if (coeffs[i] != scale * in2.coeffs[i])
	  return false;
      }

      long double c1 = this -> c;
      long double c2 = scale * in2.c;
      if ( c1 <= c2) {
	subsumes=true;
      }
      if (c2 <= c1){
	isSubsumed=true;
      }
      return true;
    } 

    return false;
  }

  bool checkRedundant( Box const & bx) {
    if (isRedundant) return true;
    if (boxSatisfies(bx)) {
      isRedundant = true;
    }
    return isRedundant;
  }

  bool isConstraintRedundant() const {
    return isRedundant;
  }

  bool  compressBox(Box const & bx, int dim, bool & hasLB, long double & lb, bool & hasUB, long double & ub) const{
    PRE(dim >= 0);
    PRE(dim < coeffs.size());
    if (isRedundant) return false;
    if (coeffs[dim] == 0.0) return false ;
    int i;
    long double b = 0.0;
    // Find a lower bound to the expression
    for (i = 0; i < coeffs.size(); ++i){
      if (i == dim) continue;
      if (coeffs[i] < 0.0){
	if (bx.hasUpperBound(i)){
	  b = b + coeffs[i] * bx.upperBound(i);
	} else {
	  return false;
	}
      } else if (coeffs[i] > 0.0){
	if (bx.hasLowerBound(i)){
	  b = b + coeffs[i] * bx.lowerBound(i);
	} else {
	  return false;
	}
      }
    }
    
    b = (c - b)/coeffs[dim];
    bool change = false;
    if (coeffs[dim] > 0.0){
      if (!hasUB){
	hasUB = true;
	ub = b;
	change = true;
      } else {
	if (ub > b) {
	  ub = b;
	  change = true;
	}
      }
      
    } else {
      assert(coeffs[dim] < 0.0);
      if (!hasLB){
	hasLB = true;
	lb = b;
	change = true;
      } else {
	if ( lb < b){
	  lb = b;
	  change = true;
	}
      }

    }
    
    return change;
  }


  void evaluateAgainstBox(vector<bool> const & fLB, vector<bool> const & fUB, 
			  vector<long double> const & lb, vector<long double> const & ub, int dim, 
			  bool & foundLB, bool & foundUB, long double & lNew, long double & uNew) const {

    
    PRE(dim >=0);
    PRE(dim < coeffs.size());

    if (isRedundant) return;

    if (coeffs[dim] == 0.0) return;
    long double b=0.0;
    int i;
    for (i = 0; i < coeffs.size(); ++i){
      if (i == dim){
	continue;
      }
      if (coeffs[i] == 0.0) continue;
      if (coeffs[i] < 0.0){
	
	  assert(fUB[i]);
	  b = b + coeffs[i] * lb[i];
	
      } else {
	assert(fLB[i]);
	b = b + coeffs[i] * ub[i];
	
      }
    }

    b = (c - b) /coeffs[dim];
    if (coeffs[dim] > 0.0){
      // xdim <= (c0 - LB(REST))/cdim
      if (!foundUB){
	foundUB = true;
	uNew = b;
      } else {
	if (b <= uNew)
	  uNew = b;
      }
    } else {
      // xdim >= (-c0 + LB(REST))/(-cdim)
      // b = (b - c0)/ (-coeffs[dim]);
      if (!foundLB){
	foundLB = true;
	lNew = b;
      } else {
	if ( b >= lNew){
	  lNew = b;
	}
      }

    }

  }


  int getIneqData(int * ind, double * val, double & coeff){
    assert(!isRedundant);
    int len = 0;
    coeff = c;
    int i;
    for (i = 0; i < coeffs.size(); ++i){
      if (coeffs[i] > EPS || coeffs[i] < -EPS ){
	ind[1+len] = 1+i;
	val[1+len] = coeffs[i];
	len++;
      }
    }

    return len;
  }

  void outputIneqToLatte (ostream & out, float latteScale) const{
    if (isRedundant) return;
    out << c*latteScale << " ";
    int i;
    for (i =0; i < coeffs.size(); ++i){
      out << -coeffs[i] * latteScale << " ";
    }
    out << endl;
  }
 
  friend ostream & operator<< ( ostream & out, Ineq const  & in){
    int i;
    if (in. isRedundant) return out;
    bool first = true;
    for (i =0; i < in.coeffs.size(); ++i){
      if (in(i) == 0.0) continue;
      if (in(i) <  0.0){
	if (in(i) == -1.0){
	  out << "-";
	} else {
	  out << in(i);
	  out << " * ";
	}
      } else {
	if (!first)
	  out << "+";
	if (in(i) != 1.0) {
	  out << in(i);
	  out << " * ";
	}
      }

      out << varData[i].getName();
      first = false;
    }
    if (first) 
      out << " 0 ";
    out << " <= " << in.c << endl;
    return out;
  }

  bool unionVars(dset_t * cl) const {
    
    bool retVal = true;

    if (isRedundant) return retVal;
    // First collect nonzero coefficients.
    
    int firstNonZero = -1;
    int i;
    for (i = 0; i < (int) coeffs.size(); ++i){
      if (coeffs[i] > EPS || coeffs[i] < -EPS ){
	if (firstNonZero < 0)
	  firstNonZero = i;
	else {
	  bool ch = doUnion(cl,firstNonZero,i);
	  if (ch){ retVal = false; }
	}
      }
    }

    return retVal;
  }

  bool hasVarInCommon(set<int> const & s) const {
    if (isRedundant) return false;
    BOOST_FOREACH(int i, s){
      if (coeffs[i] > EPS || coeffs[i] < -EPS){
	return true;
      }
    }
    return false;
  }
};
