/*---

volComp: A prototype Volume Computation Tool.

    Copyright (C) 2013  Sriram Sankaranarayanan <srirams@colorado.edu>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

---*/

#include "poly.hh"
#include <cstdio>
#include <cstdlib>
#include <unistd.h>
#include <getopt.h>

namespace po = boost::program_options;
int depth = 12;
char intlBuf[1024]; // Needed for printing
const double formFactor = 0.9;

/*-- 
  utilities for reading in information.
  calculating probabilities by MC simulation.
  --*/
int vindex = 0;
gsl_rng * r;
bool debug =false;
bool useICP = true;
bool useTerminalLP = false;
bool redundancyEliminate = true;
bool latteOutput = false ;
char latteOutputFile[200];
bool testProductDecomposition = false;
bool productDecompose=true;
bool prodDecomposeDepthBound=6;
vector<RVar> varData;

string parseOptions(int argc, char * argv[]){
  po::options_description desc("Allowed Options");
  desc.add_options()
    ("debug,v",  "Verbose output (debug)")
    ("help", "produce help message")
    ("depth,d", po::value<int>(&depth), "set the search depth (default = 12)")
    ("latte,t", po::value<string>(), "Output to latte")
    ("test1", "Test Cartesian Decomposition")
    ("input-file", po::value<string>(), "input file")
    ("product,p", "Toggle product decomposition (default = true)")
    ;

  po::positional_options_description p;
  p.add("input-file", -1);
  po::variables_map vm;
  po::store(po::command_line_parser(argc,argv).options(desc).positional(p).run(), vm);
  po::notify(vm);
  
  if (vm.count("debug")){
    debug = true;
    cout << "Setting debug to true " << endl;
  }

  cout << "Search depth is set to " << depth << endl;
  
  if (vm.count("latte") || vm.count("t")){
    latteOutput = true;
    string fName = vm["latte"].as<string>();
    latteOutputFile[191] = '\0';
    latteOutputFile[190] = '\0';
    strncpy(latteOutputFile, fName.c_str(), 190);
    cout << "LattE output will be to file " << latteOutputFile << endl;
  }

  if (vm.count("test1")){
    testProductDecomposition = true;
    debug = true;
    cout << "Testing product decomposition " << endl;
    
  }
  string inpFile = "";
  if (vm.count("input-file")){
    inpFile = vm["input-file"].as<string>();
  } else {
    cerr << "Need to specify input file. "<< endl;
    exit(1);
  }

  if (vm.count("product")){
    productDecompose = !productDecompose;
    // cout << "Product decomposition is turned on? " << productDecompose?"true":"false" << endl;
  }
  
  return inpFile;

}


RVar readVariable(ifstream & inFile){
  string name;
  char dType;
  float l;
  float u;
  vindex ++;
  inFile >> name;
  inFile >> dType;
  inFile >> l;
  inFile >> u;
  cout << "V " << name << ":" << dType << ":" << l <<", " << u << endl;
  switch (dType){
  case 'U':
    return RVar::uniformVariable(vindex, name, l,u);
    break;
  case 'G':
    return RVar::gaussianVariable(vindex, name, l, u);
    break;
  case 'I':
    return RVar::uniformIntVariable(vindex,name,ceil(l), floor(u));
  default: assert(false);
  }
}

void readConstraint(ifstream & inFile, Polytope * p){
  int i;
  Ineq e (vindex);
  double coeff;
  //cout << "here I am ! " << vindex << endl;
  for (i=0; i < vindex ; ++i){
    inFile >> coeff;
    
    e[i] = coeff;
    //cout << i << "," << coeff << endl;
  }
  inFile >> coeff;

  e.setConstant(coeff);
  //cout << e << endl;
  p -> addInequality(e);

}

void mcEstimateProbability( Polytope * p, bool bb) {
  if (bb){
    p -> computeBoundingBoxUsingLP();
    if ( p -> isEmptyPolyhedron()) {
      cout << "Warning: input polyhedron is empty!" << endl;
    }
    double pb = p -> computeBoundingBoxProbability();
    cout << "Upper bound estimate (1Box):" << pb << endl;
    double pr = p -> computeProbability(10000,true);
    cout << "pBoxEst = " << pr * pb << endl;
  } else {
    //    p -> computeBoundingBoxUsingLP();
    // double pb = p -> computeBoundingBoxProbability();
    double pr = p -> computeProbability(10000,false);
    cout << "pSimpleEst = " << pr  << endl;
  }
}


void mcBoundProbabilityRec( Polytope * p, int maxDepth, int nBranches, prob_type_t & pVal, prob_type_t & pInner){
  if (useICP){
    p -> computeBoundingBoxUsingICP();
  } else {
    p -> computeBoundingBoxUsingLP ();
  }

  if (p -> isABox()){
    prob_type_t q;
    allocNumberProb(q,0.0);
    p -> computeBoundingBoxProbability(q); 
    addAssign(pVal,q);
    addAssign(pInner, q);
    deallocNumberProb(q);
    return;
  }

  if (p -> isEmptyPolyhedron()) return;
  
  if (p -> isABox()){
    prob_type_t q;
    allocNumberProb(q,0.0);
    p -> computeBoundingBoxProbability(q); 
    addAssign(pVal,q);
    addAssign(pInner, q);
    deallocNumberProb(q);
    return;
  }

  if (maxDepth <= 0) {					
  
    prob_type_t q1;
    allocNumberProb(q1,0.0);
    p -> computeBoundingBoxProbability(q1);
    addAssign(pVal,q1);
    assignProb(q1,0.0);
    p -> computeInnerApproximation(q1);
    addAssign(pInner,q1);
    deallocNumberProb(q1);
    //    deallocNumberProb(q2);
    return; // Nothing more to do.
  }
  vector<Polytope*> decomposedResult;
  bool rVal = false;
  if (productDecompose){
    rVal = p -> cartesianDecompose(decomposedResult);
  } 
  if (!rVal){
    int dim = p -> selectBranchDimension();
    vector<Polytope*> branches;
    p -> branchAndBound(dim, nBranches, branches);
    BOOST_FOREACH(Polytope * q, branches){
      mcBoundProbabilityRec(q,maxDepth -1, nBranches, pVal,pInner);
      delete(q);
    }
  } else {
    //cout << "CARTPROD --> " << maxDepth << endl;
    prob_type_t x,y;
    allocNumberProb(x,1.0);
    allocNumberProb(y,1.0);
    prob_type_t a,b;
    allocNumberProb(a,0.0);
    allocNumberProb(b,0.0);
    BOOST_FOREACH(Polytope * pp, decomposedResult){
      assignProb(a,0.0);
      assignProb(b,0.0);
   
      if (pp -> isABox()){
	pp -> computeBoundingBoxProbability(a);
	addAssign(b,a);
      } else {
	int dim = pp -> selectBranchDimension();
	assert(dim >= 0);
	vector<Polytope*> branches;
	pp -> branchAndBound(dim, nBranches, branches);
	BOOST_FOREACH(Polytope * q, branches){
	  mcBoundProbabilityRec(q,maxDepth -1, nBranches, a,b);
	  delete(q);
	}
      }
      mulAssign(x,a);
      mulAssign(y,b);
      delete(pp);
    }
    
    
    addAssign(pVal,x);
    addAssign(pInner,y);
    deallocNumberProb(a);
    deallocNumberProb(b);
    deallocNumberProb(x);
    deallocNumberProb(y);
  }
  
  return;
}

void mcBoundProbability(Polytope * p, int bbDepth, int numBranches ){
  // Todo: consider branching strategy.
  // Todo: consider variable selection strategy: currently we select variable with largest bounds.
  
  //p -> computeBoundingBoxUsingLP();
  prob_type_t pVal, pInner;
  allocNumberProb(pVal,0.0);
  allocNumberProb(pInner,0.0);
  if (! p -> isEmptyPolyhedron())
    mcBoundProbabilityRec(p,bbDepth, numBranches,pVal,pInner);
  cout << " Probability Upper Bound : " << probPrint(pVal) << endl;
  cout << " Probability Lower Bound : "<< probPrint(pInner)<<endl;

  deallocNumberProb(pVal);
  deallocNumberProb(pInner);
  return;
}


Polytope *  readDataFile(ifstream & inFile, vector<RVar>  & vars){
  Polytope * pRet = NULL;
  char c;
  bool readVars = true;
  bool ok = true;
  while (ok && !inFile.eof()){
    inFile >> c;
    //cout << c << endl;
    switch (c){
    case 'V': {
      assert(readVars);
      RVar r = readVariable(inFile);
      vars.push_back(r);
    }
      break;
    case 'C':
      if (readVars) {
	pRet = new Polytope();
      }
      assert(pRet);
      readVars= false;
      readConstraint(inFile, pRet);
      break;
    default: 
      ok = false;
      break;
    }
  }
  return pRet;

}
// void usage(char * name){
//   cout << "Usage: " << name << " [options..] file " << endl;
//   cout << "Options are " <<endl;
//   cout << "      -d <number>  set deth of branch and bound (default = 12) " <<endl;
//   cout << "      -b <number>  set branching (default =2 )" << endl;
//   cout << "      -l           use LP solver instead of ICP " << endl;
//   cout << "      -s <number>  seed the RNG with a seed " << endl;
//   cout << "      -i           use ICP with terminal call to LP solver " << endl;
//   cout << "      -r           turn of redundancy elimination. " << endl;
//   cout << "      -t <filename> dump output to latte  " << endl;
// }

int main(int argc,  char * argv[]){
 
  int numBranches = 2;
  int seed = 30013;
  int c;
  // while ( (c = getopt(argc, argv, "hilrd:b:s:t:")) != -1) {
  //   switch(c){
  //   case 'h':
  //     usage(argv[0]); 
  //     exit(1);
  //     break;
  //   case 'l':
  //     useICP = false;
  //     break;
  //   case 'i': 
  //     useICP = true;
  //     useTerminalLP=true;
  //     break;
  //   case 'd':
  //     depth = atoi(optarg);
  //     cout << depth << endl;
  //     break;
  //   case 'r':
  //     redundancyEliminate=false;
  //     break;
  //   case 'b':
  //     numBranches = atoi(optarg);
  //     break;
  //   case 's':
  //     seed = atoi(optarg);
  //     break;
  //   case 't':
  //     latteOutput = true;
  //     strncpy(latteOutputFile, optarg,199);
  //     break;
  //   default:
  //     printf ("Usage: %s -d <depth> -b < numBranches> -s <seed for rng> -t <filename for latte output> list of files... \n ", argv[0]);
  //     exit(1);
  //     break;
  //   }
  // }

  string inputFile = parseOptions(argc,argv);

  int index;
  const gsl_rng_type * T;
  gsl_rng_env_setup();
  T = gsl_rng_default;
  r  = gsl_rng_alloc(T);
  gsl_rng_set(r,seed);
  int i;
  // const char * fName = inputFile.c_str();
  cout << "Processing file " << inputFile << endl;
  ifstream inFile;
  inFile.open(inputFile.c_str(), ios::in);
  if (!inFile){
    cerr << "Cannot open the file: " << inputFile <<endl;
    exit(1);
  }
  
  Polytope * p = readDataFile(inFile, varData);
  cout << *p << endl;

  if (latteOutput){
    ofstream latteFile;
    latteFile.open(latteOutputFile);
    if (!latteFile){
      cerr << "Cannot open file " <<latteOutputFile << " for output"<<endl;
      exit(1);
    }
    p -> outputToLatte(latteFile);
    latteFile.close();
    return 0;
  }

  if (testProductDecomposition){
    p -> testProductDecomposition();
    return 0;
  }
  
  if (p -> isEmptyPolyhedron()){
    cout << " --- without bounding box reduction --- " << endl;
    cout << "pSimpleEst = 0.0 " << endl;
    cout << " -- with bounding box reduction -- " << endl;
    cout << "pBoxEst = 0.0 " << endl;
    cout << "  -- Bounding Box upper bound computation -- " << endl;
    cout << "Depth: " << depth << endl;
    cout << "nBranches:" << numBranches << endl;
    cout << " Probability Upper Bound : " << 0.0 << endl;
    cout << " Probability Lower Bound : " << 0.0 << endl;
  } else {
    cout << " --- without bounding box reduction --- " << endl;
    mcEstimateProbability(p,false);
    cout << " --- with bounding box reduction -- " << endl;
    mcEstimateProbability(p,true);
    cout << " -- Bounding Box upper bound computation -- " << endl;
    cout << "Depth: " << depth << endl;
    cout << "nBranches:" << numBranches << endl;
    mcBoundProbability(p, depth,numBranches);
  }
  
  delete(p);
  
  gsl_rng_free(r);
  return 1;
}
