/*---

volComp: A prototype Volume Computation Tool.

    Copyright (C) 2013  Sriram Sankaranarayanan <srirams@colorado.edu>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

---*/

#include "myUtils.hh"
#include "poly.hh"
#include "unionFind.hh"
#include <cstdlib>
#include <glpk.h>
#define EPS 1E-05
extern bool debug;
bool useIPT = false;
void solveSimplexAndGetResult(glp_prob * lp, bool & feas, bool & bnded, long double & optVal){
  if (!useIPT){
    glp_smcp parm;
    glp_init_smcp(&parm);
    parm.msg_lev = GLP_MSG_OFF;
    glp_simplex(lp, &parm);
    
    feas=bnded = false;
    optVal = -12345.6;
    int stat = glp_get_status(lp);
    switch (stat){
    case GLP_OPT:
      feas = true;
      bnded = true;
      optVal = glp_get_obj_val(lp);
      break;
    case GLP_NOFEAS: 
    case GLP_INFEAS:
      feas=false;
      break;
    case GLP_UNBND:
      feas = true;
      bnded = false;
      break;
    default:
      assert (false);
    }
  } else {
    glp_iptcp p;
    glp_init_iptcp(&p);
    p.msg_lev = GLP_MSG_OFF;
    glp_interior(lp, &p);
    feas = bnded = false;
    int stat = glp_ipt_status(lp);
    switch (stat){
    case GLP_OPT:
      feas = true;
      bnded = true;
      optVal = glp_ipt_obj_val(lp);
      break;
    case GLP_NOFEAS: 
    case GLP_INFEAS:
      feas=false;
      break;
    case GLP_UNBND:
    case GLP_UNDEF:
      feas = true;
      bnded = false;
      break;
    default:
      assert (false);
    } 
  }
}

void Polytope::computeBoundingBoxUsingLP () {
  assert(!isEmpty);
  int nRows =  pruneRedundantConstraints();
  if (nRows == 0) return; // Nothing more to do.
    //assert(nRows > 0);
  glp_prob * lp = glp_create_prob();
  glp_set_prob_name(lp, "sample");
  glp_set_obj_dir(lp, GLP_MAX);
  glp_add_rows(lp, nRows);
  glp_add_cols(lp, nVars);
  
  int i;
  int * ind = new int[nVars+1];
  double * val = new double[nVars +1];
   double c;
  int len;
  for (i = 0; i < nVars; ++i){
    if (bx.hasLowerBound(i) && bx.hasUpperBound(i)){
      long double l = bx.lowerBound(i);
      long double u = bx.upperBound(i);
      if ( l +EPS < u - EPS){
	glp_set_col_bnds(lp,1+i, GLP_DB, bx.lowerBound(i), bx.upperBound(i));
      } else {
	glp_set_col_bnds(lp,1+i, GLP_FX, bx.lowerBound(i), bx.upperBound(i));
      }
    } else if (bx.hasLowerBound(i) && !bx.hasUpperBound(i)) {
      glp_set_col_bnds(lp,1+i, GLP_LO, bx.lowerBound(i), bx.lowerBound(i));
    } else if (!bx.hasLowerBound(i) && bx.hasUpperBound(i)) {
      glp_set_col_bnds(lp,1+i, GLP_UP, bx.upperBound(i), bx.upperBound(i));
    } else {
      glp_set_col_bnds(lp,1+i, GLP_FR, -1.0, 100.0);
    }
    // Clear all objective coefficients.
    glp_set_obj_coef(lp, 1+i, 0.0);
  }
  int count = 0;
  for (i=0; i < (int) ineqs.size();++i){
    if (ineqs[i].isConstraintRedundant()) continue;
    len = ineqs[i].getIneqData(ind, val, c);
    glp_set_row_bnds(lp, 1+count, GLP_UP, 0.0, c);
    glp_set_mat_row(lp, 1+count, len, ind, val);
    count ++;
  }
  assert(count == nRows);
  bool feas = true;
  bool bnded = true;
  long double optVal = -1.0;

  // Now we can compute upper and lower bounds for the variables one by one.
  for (i=0; i < nVars; ++i){
    // maximize
    glp_set_obj_coef(lp, 1+i, 1.0);
    
    solveSimplexAndGetResult( lp, feas, bnded, optVal);
    // assert(feas);
    if (!feas){
      isEmpty = true;
      return;
    }
    if (!bnded){
      assert (!bx.hasUpperBound(i));
    } else {
      assert(feas && bnded);
      bx.setUpperBound(i,optVal);
    }


    // minimize
    glp_set_obj_coef(lp, 1+i, -1.0);
    solveSimplexAndGetResult( lp, feas, bnded, optVal);
    assert(feas);
    if (!bnded){
      assert (!bx.hasLowerBound(i));
    } else {
      assert(feas && bnded);
      bx.setLowerBound(i,-optVal);
    }
    // cleanup
    glp_set_obj_coef(lp, 1+i, 0.0);
  }
  delete(ind);
  delete(val);
  glp_delete_prob(lp);
  boxCheckEmpty();
}

Polytope * copyAndRestrict(Polytope * p, int dim, bool hasLB, long double lb, bool hasUB, long double ub){
  Polytope * pNew = new Polytope(*p);
  if (hasLB && hasUB){
    assert(lb <= ub);
  }
  if (hasLB){
    pNew -> setLowerBound(dim, lb);
  } else {
    pNew -> unsetLowerBound(dim);
  }

  if (hasUB){
    pNew -> setUpperBound(dim, ub);
  } else {
    pNew -> unsetUpperBound(dim);
  }
  pNew -> boxCheckEmpty();
  return pNew;
}

void Polytope::branchAndBound(int dim, int nBranches, vector<Polytope*> & res) {
  assert(dim >= 0);
  assert(dim < nVars);
  assert(nBranches > 1);
  assert(!isEmpty);
  long double l, u;
  long double delta;
  int i,j;
  Polytope * pNew;
  bool isInteger = varData[dim].isIntegerVariable();

  if (bx.hasBothBounds(dim)){
    l = bx.lowerBound(dim);
    u = bx.upperBound(dim);
    // assert( l < u);
    delta = (u -l)/(long double) nBranches;
    if (!isInteger){
      for (i = 0; i < nBranches; ++i){
	pNew = copyAndRestrict(this, dim, true, l + (long double) i * delta, true, l + (long double ) (i+1) * delta);
	res.push_back(pNew);
      }
      assert(l + (i+1)*delta >= u);
    } else {
      if (delta <= 1.0) {
	for (i= (int)l ; i <= (int) u; ++i){
	  pNew = copyAndRestrict(this, dim, true, i, true, i);
	  res.push_back(pNew);
	}
      } else {
	long double x,y;
	for (i = 0; i < nBranches; ++i){
	  x = l + ((long double) i) * delta;
	  y = x + delta;
	  if (i > 0 && ceil(x) == floor(x)){
	    x = x +.5;
	  }
	  pNew = copyAndRestrict(this, dim, true, ceil(x), true, floor(y));
	  res.push_back(pNew);
	}
	//	assert(y >= u);
      }

    }
      
  } else if (bx.lowerBound(dim)){
    l = bx.lowerBound(dim);
    u = l + 10.0;
    delta = (u-l)/(long double) (nBranches-1);
    for (i=0; i < nBranches -1; ++i){
      pNew = copyAndRestrict(this, dim, true, l + (long double) i * delta, true, l + (long double ) (i+1) * delta);
      res.push_back(pNew);
    }
    assert(l + (i+1)*delta >= u);
    pNew = copyAndRestrict(this, dim, true, u, false, 0.0);
    //    pNew = new Polytope(*this);
    //  pNew -> setLowerBound (dim, u);
    // pNew -> unsetUpperBound(dim);
    res.push_back(pNew);
  } else if (bx.hasUpperBound(dim)){
    u = bx.upperBound(dim);
    l = u - 10.0;
    delta = (u-l)/(long double) (nBranches-1);
    for (i=0; i < nBranches -1; ++i){
      pNew = copyAndRestrict(this, dim, true, u - (i+1)* delta, true, u - i * delta);
      // pNew = new Polytope(*this);
      //  pNew -> setLowerBound(dim, u - i*delta);
      //  pNew -> setUpperBound(dim, u - (i+1) * delta);
      res.push_back(pNew);
    }
    pNew = copyAndRestrict(this, dim, false, 0.0, true, l);
    // pNew = new Polytope(*this);
    //  pNew -> setUpperBound (dim, l);
    //  pNew -> unsetLowerBound(dim);
    res.push_back(pNew);
  } else {
    if (nBranches == 2){
      pNew = copyAndRestrict(this, dim, false, 0.0, true, 0.0);
      // pNew = new Polytope (*this);
      //  pNew -> setUpperBound (dim, 0.0);
      //   pNew -> unsetLowerBound(dim);
      res.push_back(pNew);
      //      pNew = new Polytope (*this);
      // pNew -> setLowerBound (dim, 0.0);
      //  pNew -> unsetUpperBound(dim);
      pNew = copyAndRestrict(this, dim, true, 0.0, false, 0.0);
      res.push_back(pNew);
    } else {
      assert(nBranches > 2);
      pNew = copyAndRestrict(this, dim, false, 0.0, true, -10.0);
      // pNew = new Polytope (*this);
      // pNew -> setUpperBound (dim, -10.0);
      // pNew -> unsetLowerBound(dim);
      res.push_back(pNew);
      pNew = copyAndRestrict(this, dim, true, 10.0, false, 0.0);
      res.push_back(pNew);
      l = -10.0;
      u = 10.0;
      nBranches = nBranches -2;
      delta = (u -l)/(long double) nBranches;
      for (i = 0; i < nBranches; ++i){
	// l + i * delta, l + (i+1) * delta
	pNew = copyAndRestrict(this, dim, true, l+(long double) i * delta, true, l + (long double) (i+1) * delta);
	res.push_back(pNew);
    } 

    }
  } 
}

void Polytope::expandBoxAlongDimension(vector<bool> & fLB, vector<bool> & fUB, vector<long double> & lb, vector<long double> & ub, int i) const{
  bool foundLB = bx.hasLowerBound(i);
  bool foundUB = bx.hasUpperBound(i);
  long double lNew = bx.lowerBound(i);
  long double uNew = bx.upperBound(i);
  BOOST_FOREACH(Ineq const & iq, ineqs){
    iq.evaluateAgainstBox(fLB, fUB, lb, ub, i, foundLB, foundUB, lNew, uNew);
  }
  if (foundLB)
    {
      fLB[i]=true;
      lb[i] = lNew;
    } 

  if (foundUB)
    {
      fUB[i] = true;
      ub[i] = uNew;
    }
  
  
}

void Polytope::innerApproximationFromSamplePoint(vector<long double> const & pt, prob_type_t & p) const {
  assignProb(p,1.0);
  vector<bool> fLB(nVars,true);
  vector<bool> fUB(nVars,true);
  vector<long double> lb(pt);
  vector<long double> ub(pt);
  int i;
  for (i = 0; i < nVars; ++i){
    expandBoxAlongDimension(fLB, fUB, lb, ub, i);
    long double q = varData[i].computeIntervalProbability(fLB[i], lb[i], fUB[i], ub[i]);
    multiplyFloatAssign(p,q);
  }
  
  return;

}


long double Polytope::innerApproximationFromSamplePoint(vector<long double> const & pt) const{
  vector<bool> fLB(nVars,true);
  vector<bool> fUB(nVars,true);
  vector<long double> lb(pt);
  vector<long double> ub(pt);
  long double p = 1.0;
  int i;
  for (i = 0; i < nVars; ++i){
    expandBoxAlongDimension(fLB, fUB, lb, ub, i);
    p *= varData[i].computeIntervalProbability(fLB[i], lb[i], fUB[i], ub[i]);
  }
  
  return p;
}


/*-- Cartesian Decompsition of polyhedra.  


  1. Collect MSCCs of the constraint graph where xi --- xj 
     iff xi and xj occur together in some constraint.
  2. Check if it is worth decomposing by counting the number of inequalities
     in each MSCC.
  3. If at least two (or more) non trivial clusters remain, then perform the
  decomposition.

  ---*/

typedef  map< int, set<int> >::value_type vType;

bool Polytope::cartesianDecompose(vector<Polytope*> & retVec) {
  dset_t * cl = createDisjointSets(nVars);
  bool iterDone = false;

  while (!iterDone){
    iterDone = true;
    // Go through every inequality and  pairwise link the sets involved.
    BOOST_FOREACH(Ineq const & iq, ineqs){
      if (!iq.unionVars(cl))
	iterDone = false;
    }
    
  }
  // Now collect the clusters
  map<int, set<int> > clusters;
  map<int, vector<Ineq> > ineqClusters;
  collectClusters(cl,clusters);
  // Check if the decomposition is worth it
  int numNonZeroClusters = 0;

  BOOST_FOREACH( vType & myPair, clusters){
    int j = myPair.first;
    set<int> & clI = myPair.second;
    // Check if any inequality belongs to this cluster.
    int ineqCount = 0;
    BOOST_FOREACH(Ineq const & iq, ineqs){
      if (iq.hasVarInCommon(clI)) { 
	ineqCount++;
	ineqClusters[j].push_back(iq);
      }
    }
    if (ineqCount > 0) numNonZeroClusters++;
  }

  if (numNonZeroClusters <= 1){
    return false; // It is not worth decomposing.
  }

  // Otherwise, lets go for it.
  BOOST_FOREACH( vType & myPair, clusters){
    int j = myPair.first;
    set<int> & clI = myPair.second;
    map<int, vector<Ineq> >::iterator iT = ineqClusters.find(j);
    Polytope * nPolytope = new Polytope(); 
    // Prepare the bounds for the variables in the cluster.
    BOOST_FOREACH(int k, clI){
      if (bx.hasUpperBound(k)){
	nPolytope -> setUpperBound(k,bx.upperBound(k));
      }
      if (bx.hasLowerBound(k)){
	nPolytope -> setLowerBound(k,bx.lowerBound(k));
      }
    }

    if (iT != ineqClusters.end()){
      vector<Ineq> const & vecIneq = iT -> second;
      BOOST_FOREACH(Ineq const & jQ, vecIneq){
	nPolytope -> ineqs.push_back(jQ);
      }
    }
    retVec.push_back(nPolytope);
  }
  
  
  deleteDisjointSets(cl);
  return true;
}

void Polytope::testProductDecomposition()  {
  vector<Polytope*> cProdVec;
  bool rVal = cartesianDecompose(cProdVec);
  if (!rVal){
    cout << " Nothing to decompose " << endl;
  }
  int count = 1;
  BOOST_FOREACH(Polytope * p, cProdVec){
    cout << "----- Polytope # "<< count << " ------" << endl;
    cout << *p << endl;
    count ++;
  }
  cout << "Done with decomposition " << endl;
}
