/*---

volComp: A prototype Volume Computation Tool.

    Copyright (C) 2013  Sriram Sankaranarayanan <srirams@colorado.edu>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

---*/

#ifndef D__POLY__HH__
#define D__POLY__HH__

#include <vector> 
#include <ostream>
#include <map>
#include "myUtils.hh"
#include "rVar.hh"
#include "box.hh"
#include "ineq.hh"
#include "unionFind.hh"

extern gsl_rng * r;
extern bool debug;
extern bool redundancyEliminate;
extern std::vector<RVar> varData;

class Polytope {

protected:
  int nVars; // The number of variables
  //  std::vector<RVar> const & varData; // Data about the variables 
  Box bx;
  vector<Ineq> ineqs;
  bool isEmpty;

  bool checkBounds(int dim) const {
    return bx.checkBounds(dim);
  }

  void roundIntegerVariableBounds() {
    int i;
    for (i=0; i < varData.size() ; ++i){
      if (varData[i].isIntegerVariable()){
	bx.roundDimension(i);
      }
    }
    
  }

  void expandBoxAlongDimension(vector<bool> & fLB, vector<bool> & fUB, vector<long double> & lb, vector<long double> & ub, int i) const;
public:
  Polytope(): nVars(varData.size()),
	      isEmpty(false) {}
  
  bool isABox() const {
    BOOST_FOREACH(Ineq const & iq, ineqs){
      if (!iq.boxSatisfies(bx)) 
	return false;
    }
    return true;
  }
  
  
  int pruneRedundantConstraints(){
    int count = 0;
    BOOST_FOREACH(Ineq & iq, ineqs){
      if (!iq.checkRedundant(bx))
	count ++;
    }
    if (debug){
      cout << "# Irredundant: " << count << endl;
    }
    return count;
  }
  
  int numIneqs (){
    int count = 0;
    BOOST_FOREACH(Ineq & iq, ineqs){
      if (!iq.isConstraintRedundant())
	count ++;
    }
    return count;
  }

  int selectBranchDimension() const {
    // If a dimension is unbounded or half-bounded select that 
    // or else select max-width dimension
    int i;
    map<int,int> oMap;
    BOOST_FOREACH(Ineq const & iq, ineqs){
      iq.collectVarOccurrenceCounts(oMap);
    }
    
    for (i =nVars-1; i >= 0; --i) {
      if (oMap[i] == 0) continue;
      if (!bx.hasLowerBound(i) && !bx.hasUpperBound(i))
	return i;
    }
    
    for (i =0; i < nVars; ++i) {
      if (oMap[i] == 0) continue;
      if (!bx.hasLowerBound(i) || !bx.hasUpperBound(i))
	return i;
    }
    
    long double maxVal = bx.upperBound(0) - bx.lowerBound(0);
    int maxDim = -1;
    int maxOMap = -1;
    for (i = 0; i < nVars; ++i){
      if (oMap[i] == 0) continue;
      assert(bx.hasLowerBound(i) && bx.hasUpperBound(i));
      long double delta = bx.upperBound(i) - bx.lowerBound(i);
      if (maxDim == -1 || delta > maxVal || (delta == maxVal && oMap[i] > maxOMap)){
	  maxDim = i; 
	  maxVal = delta;
	  maxOMap = oMap[i];
      } else if (delta == maxVal  && oMap[i] > maxOMap){
	maxDim = i; 
	maxVal = delta;
	maxOMap = oMap[i];
      }
      

    }
    return maxDim;
  }

  bool isEmptyPolyhedron() const {
    return isEmpty;
  }

  void setUpperBound(int dim, long double u) {
    bx.setUpperBound(dim,u);
  }

  void setLowerBound(int dim, long double l)  {
    bx.setLowerBound(dim, l);
  }
  
  void unsetUpperBound(int dim){
    bx.unsetUpperBound(dim);
  }

  void unsetLowerBound(int dim){
    bx.unsetLowerBound(dim);
  }

  void addInequality(Ineq const & what){
    //if (what.checkRedundant(bx)) return;
    long double l, u;
    bool isLB =false, isUB=false;
    int dim;
    if (what.nVars() <= 0) {
      if (what.constCoeff() < 0.0){
	cerr <<" Encountered 0 < 1 constraint!! " <<endl;
	isEmpty=true;
      }
      return;
    }
    
    if (what.isBoxConstraint(dim, isLB,l, isUB,u)){
      if (isLB){
	cout << "constraint converted to lower bound " << l << endl;
	if (bx.hasUpperBound(dim)){
	  long double u1 = bx.upperBound(dim);
	  if ( u1 < l){
	    cout << "Conflict detected. " << endl;
	    isEmpty = true;
	    return;
	  }
	}
	bx.setLowerBound(dim,l);
      } else {
	assert(isUB);
	cout << "constraint converted to upper bound " << u << endl;
	if (bx.hasLowerBound(dim)){
	  long double l1 = bx.lowerBound(dim);
	  if (l1 > u ){
	    cout << "Conflict detected. " << endl;
	    isEmpty=true;
	    return;
	  }
	}

	bx.setUpperBound(dim,u);
      }
      return;
    }
    
    if (redundancyEliminate){
      BOOST_FOREACH(Ineq & iq, ineqs){
	bool subsumes=false;
	bool isSubsumed=false;
	long double scale = 0.0;
	if (iq.subsumesOrIsSubsumed(what, subsumes,isSubsumed,scale)){
	  cout << "Detected redundancy in input. " << endl;
	  if (subsumes) return; // Nothing more to be done.
	  if (isSubsumed){
	    // update the constant coefficient of iq and return
	    long double & c = iq.constCoeff();
	    c  = scale * what.constCoeff();
	    return;
	  }
	}
      }
    }

    ineqs.push_back(what);
  }

  void computeBoundingBoxProbability(prob_type_t & p) const {
    int i;
    assignProb(p,0.0);
    if ( isEmpty) return;
    assignProb(p,1.0);
    for (i=0; i < nVars; ++i){
      long double q = varData[i].computeIntervalProbability(
						 bx.hasLowerBound(i), 
						 bx.lowerBound(i), 
						 bx.hasUpperBound(i),
						 bx.upperBound(i));
      multiplyFloatAssign(p,q);
    }

  }
  
  long double computeBoundingBoxProbability()  const {
    int i;
    if ( isEmpty) return 0.0;
    long double p = 1;
    for (i=0; i < nVars; ++i){
      p *= varData[i].computeIntervalProbability(
						 bx.hasLowerBound(i), 
						 bx.lowerBound(i), 
						 bx.hasUpperBound(i),
						 bx.upperBound(i));
    }
    return p;
  }

  bool isSatisfied (vector<long double> const & pt ) const{
    vector<Ineq>::const_iterator iT;
    if (isEmpty) return false;
    for (iT = ineqs.begin(); iT != ineqs.end(); ++iT){
      if (! iT -> isSatisfied(pt)) return false;
    }
    if (debug) {
      cout << "SAT" << endl;
    }
    return true;
  }
  
  bool generateAndCheckSample ( ) const {
    if (isEmpty) return false;

    vector<long double> pt(nVars, 0.0);
    int i;
    if (debug){
      cout << "S: " ;
    }
    for (i = 0; i < nVars; ++i){
      pt[i] = varData[i].generateSample();
      if (debug){ 
	cout << ", " << pt[i];
      }
    }
    if (debug) cout << endl;
    return isSatisfied(pt);
  }

  bool generateAndCheckSampleBB() const {
    if (isEmpty) return false;

    vector<long double> pt(nVars, 0.0);
    int i;
    if (debug){
      cout << "S: " ;
    }
    for (i = 0; i < nVars; ++i){
      pt[i] = varData[i].generateSampleInRange(bx.hasLowerBound(i),bx.lowerBound(i), bx.hasUpperBound(i), bx.upperBound(i) );
      if (debug){ 
	cout << ", " << pt[i];
      }
    }
    if (debug) cout << endl;
    return isSatisfied(pt);
  }

  long double computeProbability( int nTrials, bool bb)  {
    int i;
    int s = 0;
    if (isEmpty) return 0.0;

    for (i =0; i < nTrials; ++i){
      if (bb){
	if (generateAndCheckSampleBB()) 
	  s++;
      } else {
	if (generateAndCheckSample() ) 
	  s++;
      }
    }

    cout << " # trials: " << nTrials << " and success : " << s << endl;
    return (long double) s/ (long double) nTrials;
  }

  
  
  void computeBoundingBoxUsingLP();
  
  void computeBoundingBoxUsingICP(){
    if (isEmpty) return;
    pruneRedundantConstraints();
    int i;
    int j;
    bool change = true;
    int count = 0;
    while (change && count < 10){
      change = false;
      count ++;
      for (j = 0; j < nVars; ++j){
	bool hasLB = bx.hasLowerBound(j);
	bool hasUB = bx.hasUpperBound(j);
	long double lb = hasLB? bx.lowerBound(j): -100000.0;
	long double ub = hasUB? bx.upperBound(j): 100000.0;
	if (hasLB && hasUB) assert(lb <= ub);
	
	
	for (i=0; i < (int) ineqs.size(); ++i){
	  bool rVal = ineqs[i].compressBox(bx, j, hasLB, lb, hasUB, ub);
	  if (rVal) change = true;
	  if (hasLB && hasUB && lb > ub){  isEmpty = true; return; }
	}
	
	
	if (hasLB && hasUB && lb > ub){  isEmpty = true; return; }
	if (!varData[j].isIntegerVariable() && hasLB && hasUB && lb == ub){
	  isEmpty=true;return;
	}
      
	if (hasLB){
	  bx.setLowerBound(j, lb);
	}
	if (hasUB) {
	  bx.setUpperBound(j, ub);
	}
      }
    }
    
    boxCheckEmpty();
    
  }

  long double innerApproximationFromSamplePoint(vector<long double> const & pt) const;

  void innerApproximationFromSamplePoint(vector<long double> const & pt, 
					 prob_type_t & p) const;

  void computeInnerApproximation( prob_type_t & p) const {
    assignProb(p,0.0);
    if (isEmpty){
      return;
    }
    int nTrials = 40;
    int i,j;
    int count;
    vector<long double> pt(nVars,0.0);
    for (j = 0; j < nTrials; ++j){
      for (i = 0; i < nVars; ++i){
	pt[i] = varData[i].generateSampleInRange(bx.hasLowerBound(i),bx.lowerBound(i), bx.hasUpperBound(i), bx.upperBound(i) );
      }
      if (isSatisfied(pt)){
	prob_type_t pNew;
	allocNumberProb(pNew,0.0);
	innerApproximationFromSamplePoint(pt,pNew);
	
	if (probType_geq(pNew,p))
	  assignProbProb(p,pNew);
	count ++;
	deallocNumberProb(pNew);
      }
    }

    //assignProb(p,q);
    return;
  }

  long double computeInnerApproximation() const {
    if (isEmpty) return 0.0;
    int nTrials = 2;
    int i,j;
    long double p = 0.0;
    int count;
    vector<long double> pt(nVars,0.0);
    for (j = 0; j < nTrials; ++j){
      for (i = 0; i < nVars; ++i){
	pt[i] = varData[i].generateSampleInRange(bx.hasLowerBound(i),bx.lowerBound(i), bx.hasUpperBound(i), bx.upperBound(i) );
      }
      if (isSatisfied(pt)){

	long double pNew =  innerApproximationFromSamplePoint(pt);
	if (debug){
	  cout << "Found inner sample p="<<pNew << endl;
	}
	if (pNew >= p)
	  p = pNew;
	count ++;
      }
    }
    //  cout << "C: " << count << endl;
    return p;
  }


  
  
  void branchAndBound(int dim, int nBranches, vector<Polytope*> & res);


  void boxCheckEmpty() {
    int i;
    for (i = 0; i < nVars; ++i){
      if (bx.hasUpperBound(i) && bx.hasLowerBound(i)){
	if (bx.lowerBound(i) > bx.upperBound(i)){
	  isEmpty=true;
	  return;
	}
	
      }
    }

  }



  void outputToLatte(ostream & out){
    /*-- 
      First output header
      --*/
    float latteScale = 1000.0;
    int m = numIneqs();
    for (int i=0; i < nVars; ++i){
      if (bx.hasLowerBound(i)) m++;
      if (bx.hasLowerBound(i)) m++;
    }
    out << m << " " << nVars+1 << endl;
    BOOST_FOREACH(Ineq const & iq, ineqs){
      iq.outputIneqToLatte(out,latteScale);
    }
    int j;
    // Now print variable bounds
    for (int i=0; i < nVars; ++i){
      if (bx.hasLowerBound(i)){
	out << -latteScale * bx.lowerBound(i) ;
	for (j = 0; j < nVars; ++j)
	  if (j == i)
	    out << " " << latteScale;
	  else
	    out << " " << 0;
	out << endl;
      }
      if (bx.hasUpperBound(i)){
	
	out << latteScale * bx.upperBound(i) ;
	for (j = 0; j < nVars; ++j)
	  if (j == i)
	    out << " " << -latteScale;
	  else
	    out << " " << 0;
	out << endl;
      }
    }
    
  }

  friend ostream & operator<< (ostream & out, Polytope const & p){
    vector<Ineq>::const_iterator iT;
    if (p.isEmpty){
      out << "EMPTY" << endl;
    } else {
      out << p.bx ;
      for (iT = p.ineqs.begin(); iT != p.ineqs.end(); ++iT){
	out << *iT ;
      }
    }
    return out;
  }
  bool cartesianDecompose(vector<Polytope*> & retVec);
  void testProductDecomposition() ;
};

#endif
