/*---

volComp: A prototype Volume Computation Tool.

    Copyright (C) 2013  Sriram Sankaranarayanan <srirams@colorado.edu>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

---*/
#ifndef D__RVAR__HH__
#define D__RVAR__HH__

#include "myUtils.hh"

typedef enum { UNIFORM, GAUSSIAN, UINT } distrib_type_t;

extern gsl_rng * r;

class RVar {
private:
  int id;
  string name;
  distrib_type_t dType;
  long double lb; // If variable is gaussian this represents the mean
  long double ub; // If variable is gaussian this represents the stdev.
 
public:
  RVar(int i, string n, distrib_type_t dr, long double l, long double u): id(i), name(n),dType(dr),lb(l), ub(u) {
    assert(r);
    if (dType == UNIFORM || dType == UINT) {
      assert( lb <= ub);
    }
    if (dType == UINT){
      lb = ceil(lb);
      ub = floor(ub);
    }
 
    if (dType == GAUSSIAN){
      assert (ub > 0.0);
    }
  }

  string const & getName () const {
    return name;
  }
  bool isIntegerVariable() const{
    return dType == UINT;
  }
  bool hasLowerBound () const {
    if (dType == UNIFORM || dType == UINT) return true;
    return false;
  }

  bool hasUpperBound () const {
    if (dType == UNIFORM || dType == UINT) return true;
    return false;
  }
  long double getLowerBound() const { 
    if (dType == UNIFORM || dType == UINT ) return lb;
    return 0.0;
  }
  long double getUpperBound() const{
    if (dType == UNIFORM || dType == UINT){
      return ub;
    } 
    return -1.0;
  }

  static RVar  uniformVariable(int i, string name,  long double lb, long double ub){
    assert (lb <= ub);
    return RVar(i, name, UNIFORM, lb, ub);
  }

  static RVar gaussianVariable(int i, string name, long double mean, long double sd){
    assert( sd > 0.0);
    return RVar(i,name, GAUSSIAN, mean, sd);
  }

  static RVar uniformIntVariable(int i, string name, long double lb, long double ub){
    assert(lb <= ub);
    return RVar(i, name, UINT, lb, ub);
  }
  
  long double computeIntervalProbability(bool hasLB, long double l, bool hasUB, long double u) const{
    switch(dType){
    case UNIFORM:
      if (!hasLB || l < lb) 
	l = lb;
      if (!hasUB || u > ub)
	u = ub;
      if (l > u)
	return 0.0;
      if (l == u && l == lb && u == ub){ 
	return 1.0;
      }
      
      return  (u - l)/ (ub - lb);
      break;
    case UINT:
       l = ceil(l);
       u = floor(u);
       if (!hasLB || l < lb) 
	 l = lb;
       if (!hasUB || u > ub)
	 u = ub;
       if ( l > u) return 0.0;
       return (u -l +1)/(ub - lb+1);
       break;
    case GAUSSIAN:
      if (hasLB && hasUB){
	long double p1 = gsl_cdf_gaussian_Q(l-lb,ub);
	long double p2 = gsl_cdf_gaussian_Q(u-lb,ub);
	return p1 - p2;
      } else if (hasLB) {
	return gsl_cdf_gaussian_Q(l-lb,ub);
      } else if (hasUB) {
	return gsl_cdf_gaussian_P(u-lb,ub);
      } else {
	return 1.0;
      }
      break;
    }
    assert(false);
    return 0.0;
  }

  long double generateSample() const{
    long double d;
    switch (dType){
    case UNIFORM:
      if (lb == ub) return lb;
      return gsl_ran_flat(r,lb,ub);
      break;
    case UINT:
      if (lb == ub) return lb;
      d = gsl_ran_flat(r,lb,ub+0.99);
      return floor(d); 

    case GAUSSIAN:
      return lb + gsl_ran_gaussian(r, ub);
      break;
    default:
      assert (false);
      break;
    }

    assert(false);
    return 0.0;
  }
  
  long double generateSampleInRange(bool hasLB, long double lbNew, bool hasUB, long double ubNew) const{
    long double d;
    long double sgn = 1.0;
    switch (dType){
    case UNIFORM:
      
      if (!hasLB) 
	lbNew = lb;
      if (!hasUB)
	ubNew = ub;

      lbNew = (lb < lbNew)? lbNew: lb;
      ubNew = (ub > ubNew)?ubNew: ub;
      assert( lbNew <= ubNew);
      if (lbNew == ubNew) return lbNew;
      return gsl_ran_flat(r,lbNew, ubNew);
    case UINT: 
      lbNew = ceil(lbNew);
      ubNew = floor(ubNew);
      if (!hasLB)
	lbNew = lb;
      if (!hasUB)
	ubNew = ub;
      lbNew = (lb < lbNew)? lbNew: lb;
      ubNew = (ub > ubNew)?ubNew: ub;
      assert( lbNew <= ubNew);
      if (lbNew == ubNew) return lbNew;
      d = gsl_ran_flat(r, lbNew, ubNew + 0.99);
      return floor(d);
      break;
    case GAUSSIAN:
      // This is a gaussian tail variable.
      // Shift it by the mean.
      if (hasLB && hasUB){
	lbNew -= lb;
	ubNew -= lb;
	assert (lbNew <= ubNew);
	if (lbNew == ubNew) return lbNew;
	if (lbNew <= 0.0 && ubNew >= 0.0){
	  do {
	    d = gsl_ran_gaussian(r,ub);
	  } while ( d < lbNew || d > ubNew);
	  return d + lb;
	} else {
	  if (lbNew < 0.0){
	    sgn = -1.0;
	    lbNew = -lbNew;
	    ubNew = -ubNew;
	  }
	  
	  assert(lbNew >= 0.0 && ubNew >= 0.0 && lbNew <= ubNew);
	  do {
	    d = gsl_ran_gaussian_tail(r, lbNew, ub);
	  } while ( d < lbNew || d > ubNew);
	  return (sgn * d) + lb;
	}
      } else if (hasLB){
	lbNew -= lb;
	if (lbNew <= 0.0){
	  do {
	    d = gsl_ran_gaussian(r, ub);
	  } while ( d < lbNew);
	  return d + lb;
	} else {
	  return lb + gsl_ran_gaussian_tail(r, lbNew, ub);
	}

      } else if (hasUB){
	ubNew -= lb;
	if (ubNew < 0.0){
	  return lb - gsl_ran_gaussian_tail(r, -ubNew, ub);
	} else { 
	  do {
	    d = gsl_ran_gaussian(r,ub);
	  } while (d > ubNew);
	  return d + lb;
	} 
      } else {
	return lb + gsl_ran_gaussian(r, ub);
      }
      
      
    default:
      assert(false);
    }
    return 0.0;
  }
  
};
#endif
